# Python Libraries
from __future__ import print_function
import os
import sys
import pysam
import shutil
import pickle
from ctypes import *
from collections import *
import natsort as natsort_ob

# Local Libraries
import numpy
numpy.seterr(divide='ignore', invalid='ignore')
import matplotlib
from matplotlib import colors
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.ticker import MaxNLocator


# Distal Libraries
from rgt.SequenceSet import SequenceSet
from rgt.viz.plotTools import output_array
from rgt.GenomicRegion import GenomicRegion
from RNADNABindingSet import RNADNABindingSet
from rgt.GenomicRegionSet import GenomicRegionSet
from rgt.motifanalysis.Statistics import multiple_test_correction
from rgt.Util import SequenceType, Html, ConfigurationFile, GenomeData, Library_path

# Color code for all analysis
target_color = "mediumblue"
nontarget_color = "darkgrey"
sig_color = "powderblue"

####################################################################################
####################################################################################

def print2(summary, string):
    """ Show the message on the console and also save in summary. """
    print(string)
    summary.append(string)


def output_summary(summary, directory, filename):
    """Save the summary log file into the defined directory"""
    pd = os.path.join(dir, directory)
    try:
        os.stat(pd)
    except:
        os.mkdir(pd)
    if summary:
        with open(os.path.join(pd, filename), 'w') as f:
            print("********* RGT Triplex: Summary information *********",
                  file=f)
            for s in summary:
                print(s, file=f)


def check_dir(path):
    """Check the availability of the given directory and creat it"""
    try: os.stat(path)
    except:
        try: os.mkdir(path)
        except: pass


def try_int(s):
    "Convert to integer if possible."
    try:
        return int(s)
    except:
        return s


def natsort_key(s):
    "Used internally to get a tuple by which s is sorted."
    import re
    return map(try_int, re.findall(r'(\d+|\D+)', s))


def natcmp(a, b):
    "Natural string comparison, case sensitive."
    return cmp(natsort_key(a), natsort_key(b))


def natcasecmp(a, b):
    "Natural string comparison, ignores case."
    return natcmp(a.lower(), b.lower())


def natsort(seq, cmp=natcmp):
    "In-place natural string sort."
    seq.sort(cmp)


def natsorted(seq, cmp=natcmp):
    "Returns a copy of seq, sorted by natural string sort."
    import copy
    temp = copy.copy(seq)
    natsort(temp, cmp)
    return temp


def list_all_index(path, link_d=None):
    """Creat an 'index.html' in the defined directory """

    dirname = os.path.basename(path)

    if link_d:
        pass
    else:
        link_d = {"List": "index.html"}

    html = Html(name="Directory: " + dirname, links_dict=link_d,
                fig_rpath="./style", fig_dir=os.path.join(path, "style"),
                RGT_header=False, other_logo="TDF", homepage="../index.html")

    html.add_heading("All experiments in: " + dirname + "/")

    data_table = []
    type_list = 'sssssssssssssssssss'
    col_size_list = [20] * 20
    c = 0

    header_list = ["No.", "Experiments", "RNA", "Closest genes",
                   "Exon", "Length", "Expression*",
                   "Norm DBS*",
                   "Norm DBD*",  "No sig. DBD",
                   "Organism", "Target region",
                   "Rank*"]

    profile_f = open(os.path.join(path, "profile.txt"), 'r')
    profile = {}
    for line in profile_f:
        line = line.strip()
        line = line.split("\t")
        if line[0] == "Experiment": continue
        elif len(line) > 5: profile[line[0]] = line[1:]
    profile_f.close()

    # sig_list = []

    for i, exp in enumerate(profile.keys()):
        c += 1
        if profile[exp][10] == "-":
            new_line = [str(c), exp, profile[exp][0]]
        else:
            new_line = [str(c),
                        '<a href="' + os.path.join(exp, "index.html") + \
                        '">' + exp + "</a>", profile[exp][0]]
        new_line += [ profile[exp][12],#3 close genes
                      profile[exp][1], #4 exon
                      profile[exp][2], #5 length
                      profile[exp][13] ]#6 exp

        if float(profile[exp][11]) < 0.05:
            new_line += [ profile[exp][6], #7 norm DBS
                          profile[exp][8], #8 norm DBD
                          profile[exp][9]] #9 sig DBD
                          # profile[exp][10], #10 Top DBD
                          # "<font color=\"red\">" + \
                          # profile[exp][11] + "</font>"]
            # sig_list.append(True)
        else:
            new_line += [str(0),  # 7 norm DBS
                         str(0),  # 8 norm DBD
                         profile[exp][9]]  # 9 sig DBD
                         # profile[exp][10],  # 10 Top DBD
                         # profile[exp][11]]
            # sig_list.append(False)

        new_line += [ profile[exp][4], profile[exp][5] ]

        data_table.append(new_line)

    rank_dbd = len(data_table) - rank_array([float(x[8]) for x in data_table])
    rank_dbs = len(data_table) - rank_array([float(x[7]) for x in data_table])

    rank_exp = len(data_table) - rank_array([0 if x[6] == "n.a." else float(x[6]) for x in data_table ])

    rank_sum = [x + y + z for x, y, z  in zip(rank_dbd, rank_dbs, rank_exp)]

    nd = [ d + [str(rank_sum[i])] for i, d in enumerate(data_table) ]

    nd = natsort_ob.natsorted(nd, key=lambda x: x[-1])
    html.add_zebra_table(header_list, col_size_list, type_list, nd,
                         align=10, cell_align="left", sortable=True)

    html.add_fixed_rank_sortable()
    html.write(os.path.join(path, "index.html"))


def revise_index(root):
    "Revise other index.html in the same project"

    dir_list = {}
    plist = {}
    for item in os.listdir(root):
        h = os.path.join(root, item, "index.html")
        pro = os.path.join(root, item, "profile.txt")
        if os.path.isfile(pro):
            dir_list[os.path.basename(item)] = "../" + item + "/index.html"
            plist[os.path.basename(item)] = h
    dir_list = OrderedDict(sorted(dir_list.items()))
    # print(dir_list)
    for d, p in plist.iteritems():
        list_all_index(path=os.path.dirname(p),
                       link_d=dir_list)

def update_profile(dirpath, expression, name_list=None):
    header_list = ["Experiment", "RNA_names", "Tag", "Organism", "Target_region", "No_sig_DBDs",
                   "Top_DBD", "p-value", "closest_genes"]
    profiles = []
    pro = os.path.join(dirpath, "profile.txt")
    if not os.path.isfile(pro):
        print("There is no profile.txt in this directory.")
        return

    if expression:
        gene_exp = {}
        with open(expression) as f:
            for line in f:
                l = line.strip().split()
                gene_exp[l[0].partition(".")[0]] = l[1]
        profile_temp = []
        with open(pro) as f:
            for line in f:
                if not line.startswith("Experiment"):
                    l = line.strip().split("\t")
                    if len(l) == 12:
                        profile_temp.append(l + [gene_exp[l[0]]])
        with open(pro, "w") as f:
            h = ["Experiment", "RNA_names", "exon", "length",
                 "Tag", "Organism", "Target_region",
                 "Norm_DBS", "Norm_DBS_on_sig_DBD",
                 "Norm_DBD", "No_sig_DBDs", "Top_DBD",
                 "p-value", "closest_genes", "expression"]
            print("\t".join(h), file=f)
            for line in profile_temp:
                print("\t".join(line), file=f)
    if name_list:
        pass
    else:
        for item in os.listdir(dirpath):
            stat = os.path.join(dirpath, item, "stat.txt")
            summary = os.path.join(dirpath, item, "summary.txt")
            if os.path.isfile(stat) and os.path.isfile(summary):
                with open(stat) as f:
                    for line in f:
                        l = line.strip().split()
                        if l[0] == "name":
                            each_name = l[1]
                            each_tag = l[1].split("_")[-1]
                        if l[0] == "genome": each_organism = l[1]
                        if l[0] == "DBD_sig": each_DBD_sig = l[1]
                        if l[0] == "p_value": each_p_value = l[1]
                with open(summary) as g:
                    for line in g:
                        if "rgt-TDF" in line and " -de " in line:
                            l = line.strip().split()
                            each_target_region = os.path.basename(l[l.index("-de") + 1])
                profiles.append([item,each_name,each_tag,each_organism,each_target_region,
                                 each_DBD_sig,"n.a.",each_p_value,"-"])

    with open(pro, "w") as f:
        print("\t".join(header_list), file=f)
        for line in profiles:
            print("\t".join(line), file=f)


def gen_heatmap(path):
    """Generate the heatmap to show the sig RNA in among conditions"""

    def fmt(x, pos):
        # a, b = '{:.0e}'.format(x).split('e')
        # b = int(b)
        # return r'${} \times 10^{{{}}}$'.format(a, b)
        a = -numpy.log10(x)
        return '{:.0f}'.format(a)

    matrix = OrderedDict()
    rnas = []
    for item in os.listdir(path):
        # print(item)
        # print(os.path.isdir(os.path.join(path,item)))
        if not os.path.isdir(os.path.join(path, item)): continue
        if item == "style": continue
        # if item == "index.html": continue
        matrix[item] = {}
        pro = os.path.join(path, item, "profile.txt")
        with open(pro) as f:
            for line in f:
                line = line.strip().split("\t")
                if line[0] == "Experiment": continue
                if line[6] == "-": continue
                matrix[item][line[0]] = float(line[7])
                rnas.append(line[0])
    rnas = list(set(rnas))
    # rnas.sort()

    # Convert into array
    ar = []
    exps = natsorted(matrix.keys())
    rnas = natsorted(rnas)
    # print(exps)
    for exp in exps:
        row = []
        for rna in rnas:
            try:
                row.append(matrix[exp][rna])
            except:
                row.append(1)
        ar.append(row)
    ar = numpy.array(ar)
    ar = numpy.transpose(ar)

    # print(ar.shape)
    data = ar[~numpy.all(ar == 1, axis=1)]
    # print(data.shape)


    fig = plt.figure(figsize=(len(matrix.keys()) * 1.5, len(rnas) * 2.5))
    # fig = plt.figure()
    # ax1 = fig.add_axes([0.09,0.2,0.2,0.6])
    # Y = sch.linkage(data, method='single')
    # Z1 = sch.dendrogram(Y, orientation='right')
    # ax1.set_xticks([])
    # ax1.set_yticks([])

    # Compute and plot second dendrogram.
    # ax2 = fig.add_axes([0.3, 1.1, 0.55,0.04])
    # Y = sch.linkage(data.T, method='single')
    # Z2 = sch.dendrogram(Y)
    # ax2.set_xticks([])
    # ax2.set_yticks([])
    bounds = []
    for n in range(-8, 0):
        # print(10**n)
        bounds.append(10 ** n)
    norm = colors.BoundaryNorm(bounds, plt.cm.YlOrRd_r.N)

    # Plot distance matrix.
    axmatrix = fig.add_axes([0.1, 0.2, 0.8, 0.6])
    # axmatrix = fig.add_axes()
    # idx1 = Z1['leaves']
    # idx2 = Z2['leaves']
    # data = data[idx1,:]
    # data = data[:,idx2]
    im = axmatrix.matshow(data, aspect='auto', origin='lower', cmap=plt.cm.YlOrRd_r, norm=norm)

    axmatrix.set_xticks(range(data.shape[1]))
    axmatrix.set_xticklabels(exps, minor=False, ha="left")
    axmatrix.xaxis.set_label_position('top')
    axmatrix.xaxis.tick_top()
    plt.xticks(rotation=40, fontsize=10)

    axmatrix.set_yticks(range(data.shape[0]))
    axmatrix.set_yticklabels(rnas, minor=False)
    # axmatrix.set_yticklabels( [ rnas[i] for i in idx1 ], minor=False)
    axmatrix.yaxis.set_label_position('right')
    axmatrix.yaxis.tick_right()
    plt.yticks(rotation=0, fontsize=10)
    # axmatrix.tight_layout()

    # Plot colorbar.
    axcolor = fig.add_axes([0.1, 0.1, 0.8, 0.02])

    plt.colorbar(im, cax=axcolor, orientation='horizontal', norm=norm,
                 boundaries=bounds, ticks=bounds, format=matplotlib.ticker.FuncFormatter(fmt))
    axcolor.set_xlabel('p value (-log10)')
    lmats = ar.tolist()
    for i, r in enumerate(rnas):
        lmats[i] = [r] + [str(x) for x in lmats[i]]
    lmats = [["p-value"] + exps] + lmats

    output_array(array=lmats, directory=path, folder="", filename='matrix_p.txt')
    # os.path.join(path,'matrix_p.txt'), lmats, delimiter='\t')
    try:
        fig.savefig(os.path.join(path, 'condition_lncRNA_dendrogram.png'))
        fig.savefig(os.path.join(path, 'condition_lncRNA_dendrogram.pdf'), format="pdf")
    except:
        pass


def generate_rna_exp_pv_table(root, multi_corr=True):
    "Generate p value table for Experiments vs RNA in the same project"

    nested_dict = lambda: defaultdict(nested_dict)
    # nested_dict = lambda: defaultdict(lambda: 'n.a.')

    data = nested_dict()
    rnas = []

    for item in os.listdir(root):
        pro = os.path.join(root, item, "profile.txt")
        if os.path.isfile(pro):
            with open(pro) as f:
                for line in f:
                    if line.startswith("Experiment"):
                        continue
                    else:
                        line = line.strip().split("\t")
                        data[item][line[0]] = float(line[7])
                        rnas.append(line[0])

    exp_list = sorted(data.keys())
    rnas = sorted(list(set(rnas)))

    pvs = []
    for rna in rnas:
        for exp in exp_list:
            if data[exp][rna]: pvs.append(data[exp][rna])
    if multi_corr:
        reject, pvals_corrected = multiple_test_correction(pvs, alpha=0.05, method='indep')
    else:
        pvals_corrected = pvs

    with open(os.path.join(root, "table_exp_rna_pv.txt"), "w") as t:
        print("\t".join(["RNA_ID"] + exp_list), file=t)
        i = 0
        for rna in rnas:
            newline = [rna]
            for exp in exp_list:
                if data[exp][rna]:
                    newline.append(str(pvals_corrected[i]))
                    i += 1
                else:
                    newline.append("n.a.")
            print("\t".join(newline), file=t)


def value2str(value):
    if (isinstance(value,str)):
        try: value = float(value)
        except: return value
    if value == 0: return "0"
    if(isinstance(value,int)): return str(value)
    elif(isinstance(value,float)):
        if abs(value) >= 1000: 
            try: r = "{}".format(int(value))
            except: r = "Inf"
        elif 1000 > abs(value) > 10: r = "{:.1f}".format(value)
        elif 10 > abs(value) >= 1: r = "{:.2f}".format(value)
        elif 1 > abs(value) >= 0.05: r = "{:.2f}".format(value)
        elif 0.05 > abs(value) > 0.0001: r = "{:.4f}".format(value)
        else: r = "{:.1e}".format(value)
        return r


def uniq(seq):
    seen = set()
    seen_add = seen.add
    return [ x for x in seq if not (x in seen or seen_add(x))]


def random_each(input):
    """Return the counts of DNA Binding sites with randomization
    For multiprocessing. 
    Input contains:
    0       1               2                3     4              5          6                    
    str(i), self.rna_fasta, self.dna_region, temp, self.organism, self.rbss, str(marks.count(i)),
    number, rna,            region,          temp, organism,      rbss,      number of mark

    7  8  9  10  11  12  13  14  15          16                 17
    l, e, c, fr, fm, of, mf, rm, filter_bed, self.genome_path,  par
    """
    import sys
    # Filter BED file
    if input[15]:
        random = input[2].random_regions(organism=input[4], multiply_factor=1,
                                         overlap_result=True, overlap_input=True,
                                         chrom_X=True, chrom_M=False, filter_path=input[15])
    else:
        random = input[2].random_regions(organism=input[4], multiply_factor=1,
                                         overlap_result=True, overlap_input=True,
                                         chrom_X=True, chrom_M=False)
    
    txp = find_triplex(rna_fasta=input[1], dna_region=random, temp=input[3],
                       organism=input[4], prefix=str(input[0]), remove_temp=True,
                       l=int(input[7]), e=int(input[8]), c=input[9], fr=input[10],
                       fm=input[11], of=input[12], mf=input[13], rm=input[14],
                       par=input[17], genome_path=input[16],
                       dna_fine_posi=False)

    txp.merge_rbs(rbss=input[5], rm_duplicate=True)

    txpf = find_triplex(rna_fasta=input[1], dna_region=random, temp=input[3], 
                       organism=input[4], prefix=str(input[0]), remove_temp=True, 
                       l=int(input[7]), e=int(input[8]),  c=input[9], fr=input[10], 
                       fm=input[11], of=input[12], mf=input[13], rm=input[14], 
                       par=input[17], genome_path=input[16],
                       dna_fine_posi=True)

    txpf.merge_rbs(rbss=input[5], rm_duplicate=True)
    sys.stdout.flush()
    print("".join(["="]*int(input[6])), end="")

    return [ [len(tr) for tr in txp.merged_dict.values() ], [len(dbss) for dbss in txpf.merged_dict.values()] ]


def get_sequence(dir, filename, regions, genome_path):
    """
    Fetch sequence into FASTA file according to the given BED file
    """
    genome = pysam.Fastafile(genome_path)
    with open(os.path.join(dir, filename), 'w') as output:
        for region in regions:
            if "_" not in region.chrom:
                print(">"+ region.toString(), file=output)
                print(genome.fetch(region.chrom, max(0, region.initial), region.final), file=output)


def find_triplex(rna_fasta, dna_region, temp, organism, l, e, dna_fine_posi, genome_path, prefix="", remove_temp=False, 
                 c=None, fr=None, fm=None, of=None, mf=None, rm=None, par=""):
    """Given a GenomicRegionSet to run Triplexator and return the RNADNABindingSet"""
    
    # Generate FASTA 
    get_sequence(dir=temp, filename="dna_"+prefix+".fa", regions=dna_region, genome_path=genome_path)

    # Triplexator
    run_triplexator(ss=rna_fasta, ds=os.path.join(temp,"dna_"+prefix+".fa"), 
                    output=os.path.join(temp, "dna_"+prefix+".txp"), 
                    l=l, e=e, c=c, fr=fr, fm=fm, of=of, mf=mf, rm=rm, par=par)
    # Read txp
    txp = RNADNABindingSet("dna")
    txp.read_txp(os.path.join(temp, "dna_"+prefix+".txp"), dna_fine_posi=dna_fine_posi)
    txp.remove_duplicates()

    if remove_temp:
        os.remove(os.path.join(temp,"dna_"+prefix+".fa"))
        os.remove(os.path.join(temp,"dna_"+prefix+".txp"))

    return txp

def run_triplexator(ss, ds, output, l=None, e=None, c=None, fr=None, fm=None, of=None, mf=None, rm=None, par=""):
    """Perform Triplexator"""
    #triplexator_path = check_triplexator_path()
    # triplexator -ss -ds -l 15 -e 20 -c 2 -fr off -fm 0 -of 1 -rm
    triclass = Library_path()
    triplex_lib_path = triclass.get_triplexator()
    triplex_lib  = cdll.LoadLibrary(triplex_lib_path)

    arguments = ""
    if ss: arguments += "-ss "+ss+" "
    if ds: arguments += "-ds "+ds+" "
    if l: arguments += "-l "+str(l)+" "
    if e: arguments += "-e "+str(e)+" "
    if c: arguments += "-c "+str(c)+" "
    if fr: arguments += "-fr "+fr+" "
    if fm: arguments += "-fm "+str(fm)+" "
    if of: arguments += "-of "+str(of)+" "
    if mf: arguments += "-mf "
    if rm: arguments += "-rm "+str(rm)+" "
    # arguments += "--bit-parallel "
    if par != "":
        par = par.replace('_'," ")
        par = "-" + par
        arguments += par+" "
    
    arguments += "-o "+ os.path.basename(output) + " -od " + os.path.dirname(output)

    arg_strings  = arguments.split(' ')
    arg_ptr      = (c_char_p * (len(arg_strings) + 1))()

    arg_ptr[0] = "triplexator"  # to simulate calling from cmd line
    for i, s in enumerate(arg_strings):
        arg_ptr[i + 1] = s
    
    triplex_lib.pyTriplexator(len(arg_strings) + 1, arg_ptr)
    os.remove(os.path.join(output + ".summary"))
    os.remove(os.path.join(output + ".log"))


def read_ac(path, cut_off, rnalen):
    """Read the RNA accessibility file and output its positions and values

    The file should be a simple table with two columns:
    The first column is the position and the second one is the value
    '#' will be skipped

    """
    access = []
    with open(path) as f:
        i = 0
        while i < rnalen:
            for line in f:
                line = line.split()
                if not line: continue
                elif line[0][0] == "#": continue
                elif len(line) < 2: continue
                else:
                    v = line[1]
                    if v == "NA": 
                        access.append(0)
                    else: 
                        try: v = 2**(-float(v))
                        except: continue
                        if v >= cut_off:
                            access.append(1)
                        else:
                            access.append(0)
                    i += 1
    return access

def region_link_internet(organism, region):
    ani = None
    if organism == "hg19":
        ani = "human"
    elif organism == "hg38":
        ani = "human"
    elif organism == "mm9":
        ani = "mouse"
    if ani:
        region_link = "".join(['<a href="http://genome.ucsc.edu/cgi-bin/hgTracks?db=', organism,
                               "&position=", region.chrom, "%3A", str(region.initial), "-",
                               str(region.final), '" style="text-align:left" target="_blank">',
                               region.toString(space=True), '</a>'])
    else:
        if organism == "tair10":
            region_link = "".join(
                ['<a href="http://tairvm17.tacc.utexas.edu/cgi-bin/gb2/gbrowse/arabidopsis/?name=',
                 region.chrom, "%3A", str(region.initial), "..", str(region.final),
                 '" target="_blank">',
                 region.toString(space=True), '</a>'])
        else:
            region_link = region.toString(space=True)
    return region_link


def split_gene_name(gene_name, org):
    
    if gene_name == None:
        return ""
    if gene_name[0:2] == "chr":
        return gene_name

    if org=="hg19": ani = "human"
    elif org=="hg38": ani = "human"
    elif org=="mm9": ani = "mouse"
    else: ani = None

    if not ani:
        if org == "tair10":
            p1 = "".join(['<a href="https://www.arabidopsis.org/servlets/TairObject?name=', gene_name,
                          '&type=locus" target="_blank">', gene_name, '</a>' ])
            return p1
        else:
            return gene_name
    else:    
        p1 = '<a href="http://genome.ucsc.edu/cgi-bin/hgTracks?org='+ani+\
             "&db="+org+"&singleSearch=knownCanonical&position="
        p2 = '" style="text-align:left" target="_blank">'
        p3 = '</a>'

        if ":" in gene_name:
            genes = gene_name.split(":")
            genes = list(set(genes))
            result = []
            c = 0
            for i, g in enumerate(genes):
                if "(" in g:
                    d = g.partition('(')[2].partition(')')[0]
                    g = g.partition('(')[0]
                    if "-" in d:
                        result.insert(0, p1+g+p2+g+p3+"("+d+")")
                    else:
                        result.append(p1+g+p2+g+p3+"("+d+")")
                else:
                    c += 1
                    if c < 6:
                        result.append(p1 + g + p2 + g + p3)

            result = ",".join(result)

        elif gene_name == ".":
            result = "none"

        else:
            if "(" in gene_name:
                d = gene_name.partition('(')[2].partition(')')[0]
                g = gene_name.partition('(')[0]
                result = p1+g+p2+g+p3+"("+d+")"
            else:
                result = p1+gene_name+p2+gene_name+p3
        
        return result


def lineplot(txp, rnalen, rnaname, dirp, sig_region, cut_off, log, ylabel, linelabel, 
             filename, ac=None, showpa=False, exons=None):
    # Plotting
    f, ax = plt.subplots(1, 1, dpi=300, figsize=(6,4))
    
    # Extract data points
    x = range(rnalen)
    #print(rnalen)
    if log:
        all_y = [1] * rnalen
        p_y = [1] * rnalen
        a_y = [1] * rnalen
    else:
        all_y = [0] * rnalen
        p_y = [0] * rnalen
        a_y = [0] * rnalen

    txp.remove_duplicates_by_dbs()
    for rd in txp:
        #print(str(rd.rna.initial), str(rd.rna.final))
        if rd.rna.orientation == "P":
            for i in range(rd.rna.initial, rd.rna.final):
                p_y[i] += 1
                all_y[i] += 1
        if rd.rna.orientation == "A":
            for i in range(rd.rna.initial, rd.rna.final):
                a_y[i] += 1
                all_y[i] += 1
    # Log
    if log:
        all_y = numpy.log(all_y)
        p_y = numpy.log(p_y)
        a_y = numpy.log(a_y)
        max_y = max(all_y)+0.5
        min_y = 1
        ylabel += "(log10)"
    else:
        max_y = float(max(all_y) * 1.1)
        min_y = 0

    if ac:
        min_y = float(max_y*(-0.09))
    
    
    # Plotting
    for rbs in sig_region:
        rect = patches.Rectangle(xy=(rbs.initial,0), width=len(rbs), height=max_y, facecolor=sig_color, 
                                 edgecolor="none", alpha=0.5, lw=None, label="Significant DBD")
        ax.add_patch(rect)
    
    lw = 1.5
    if showpa:
        ax.plot(x, all_y, color=target_color, alpha=1, lw=lw, label="Parallel + Anti-parallel")
        ax.plot(x, p_y, color="purple", alpha=1, lw=lw, label="Parallel")
        ax.plot(x, a_y, color="dimgrey", alpha=.8, lw=lw, label="Anti-parallel")
    else:
        ax.plot(x, all_y, color="mediumblue", alpha=1, lw=lw, label=linelabel)

    # RNA accessbility
    if ac:
        n_value = read_ac(ac, cut_off, rnalen=rnalen)
        drawing = False
        for i in x:
            if n_value[i] > 0:
                if drawing:
                    continue
                else:
                    last_i = i
                    drawing = True
            elif drawing:
                pac = ax.add_patch(patches.Rectangle((last_i, min_y), i-last_i, -min_y,
                                   hatch='///', fill=False, snap=False, linewidth=0, label="RNA accessibility"))
                drawing = False
            else:
                continue

    # Legend
    handles, labels = ax.get_legend_handles_labels()
    legend_h = []
    legend_l = []
    for uniqlabel in uniq(labels):
        legend_h.append(handles[labels.index(uniqlabel)])
        legend_l.append(uniqlabel)
    ax.legend(legend_h, legend_l, 
              bbox_to_anchor=(0., 1.02, 1., .102), loc=2, mode="expand", borderaxespad=0., 
              prop={'size':9}, ncol=3)

    # XY axis
    ax.set_xlim(left=0, right=rnalen )
    ax.set_ylim( [min_y, max_y] ) 
    for tick in ax.xaxis.get_major_ticks(): tick.label.set_fontsize(9) 
    for tick in ax.yaxis.get_major_ticks(): tick.label.set_fontsize(9) 
    ax.set_xlabel(rnaname+" sequence (bp)", fontsize=9)
    
    ax.set_ylabel(ylabel,fontsize=9, rotation=90)
    
    if None:
        if exons and len(exons) > 1:
            w = 0
            i = 0
            h = (max_y - min_y)*0.02

            for exon in exons:
                l = abs(exon[2] - exon[1])
                
                #print([i,l,w])
                #ax.axvline(x=w, color="gray", alpha=0.5, zorder=100)
                if i % 2 == 0:
                    rect = matplotlib.patches.Rectangle((w,max_y-h),l,h, color="moccasin")
                else:
                    rect = matplotlib.patches.Rectangle((w,max_y-h),l,h, color="gold")
                ax.add_patch(rect)
                i += 1
                w += l
            ax.text(rnalen*0.01, max_y-2*h, "exon boundaries", fontsize=5, color='black')

    f.tight_layout(pad=1.08, h_pad=None, w_pad=None)

    f.savefig(os.path.join(dirp, filename), facecolor='w', edgecolor='w',  
              bbox_extra_artists=(plt.gci()), bbox_inches='tight', dpi=300)
    # PDF
    for tick in ax.xaxis.get_major_ticks():
        tick.label.set_fontsize(12) 
    for tick in ax.yaxis.get_major_ticks():
        tick.label.set_fontsize(12)
    ax.xaxis.label.set_size(14)
    ax.yaxis.label.set_size(14) 
    ax.legend(legend_h, legend_l, 
              bbox_to_anchor=(0., 1.02, 1., .102), loc=2, mode="expand", borderaxespad=0., 
              prop={'size':12}, ncol=3)
    pp = PdfPages(os.path.splitext(os.path.join(dirp,filename))[0] +'.pdf')
    pp.savefig(f,  bbox_inches='tight') # bbox_extra_artists=(plt.gci()),
    pp.close()

def load_dump(path, filename):
    file = open(os.path.join(path,filename),'r')
    object = pickle.load(file)
    file.close()
    print("\tLoading from file: "+filename)
    return object

def dump(object, path, filename):
    file = open(os.path.join(path,filename),'wb')
    pickle.dump(object,file)
    file.close()
    print("\tDump to file: "+filename)

def check_triplexator_path():
    try:
        cf = ConfigurationFile()
        with open(os.path.join(cf.data_dir,"triplexator_path.txt")) as f:
            l = f.readline()
            l = l.strip("\n")
            return l
    except:
        print("Please define the path to Triplexator by command: rgt-TDF triplexator -path <PATH>")
        sys.exit(1)
    
def rna_associated_gene(rna_regions, name, organism):
    if rna_regions:
        s = [ rna_regions[0][0], min([e[1] for e in rna_regions]), 
              max([e[2] for e in rna_regions]), rna_regions[0][3] ]
        g = GenomicRegionSet("RNA associated genes")
        g.add( GenomicRegion(chrom=s[0], initial=s[1], final=s[2], name=name, orientation=s[3]) )
        asso_genes = g.gene_association(organism=organism, promoterLength=1000, show_dis=True)

        genes = asso_genes[0].name.split(":")
        closest_genes = []
        for n in genes:
            if name not in n: closest_genes.append(n)
        closest_genes = set(closest_genes)

        if len(closest_genes) == 0:
            return "."
        else:
            return ":".join(closest_genes)
    else:
        return "."

def rank_array(a):
    try:
        a = numpy.array(a)
    except:
        a = numpy.array([float(b) for b in a])
    sa = numpy.searchsorted(numpy.sort(a), a)
    return sa

def dbd_regions(exons, sig_region, rna_name, output,out_file=False, temp=None, fasta=True):
    """Generate the BED file of significant DBD regions and FASTA file of the sequences"""
    if len(sig_region) == 0:
        return
    #print(self.rna_regions)
    if not exons:
        pass
    else:
        dbd = GenomicRegionSet("DBD")
        dbdmap = {}
        if len(exons) == 1:
            print("## Warning: No information of exons in the given RNA sequence, the DBD position may be problematic. ")
        for rbs in sig_region:
            loop = True

            if exons[0][3] == "-":
                while loop:
                    cf = 0
                    for exon in exons:
                        #print(exon)

                        l = abs(exon[2] - exon[1])
                        tail = cf + l

                        if cf <= rbs.initial <=  tail:
                            dbdstart = exon[2] - rbs.initial + cf
                            
                            if rbs.final <= tail: 
                                #print("1")
                                dbdend = exon[2] - rbs.final + cf
                                if dbdstart > dbdend: dbdstart, dbdend = dbdend, dbdstart
                                dbd.add( GenomicRegion(chrom=exons[0][0], 
                                                       initial=dbdstart, final=dbdend, 
                                                       orientation=exons[0][3], 
                                                       name=str(rbs.initial)+"-"+str(rbs.final) ) )
                                dbdmap[str(rbs)] = dbd[-1].toString() + " strand:-"
                                loop = False
                                break
                            elif rbs.final > tail:

                                subtract = l + cf - rbs.initial
                                #print("2")
                                #print("Subtract: "+str(subtract))
                                if dbdstart > exon[1]: dbdstart, exon[1] = exon[1], dbdstart
                                dbd.add( GenomicRegion(chrom=exons[0][0], 
                                                       initial=dbdstart, final=exon[1], 
                                                       orientation=exons[0][3], 
                                                       name=str(rbs.initial)+"-"+str(rbs.initial+subtract)+"_split1" ) )
                        
                        elif rbs.initial < cf and rbs.final <= tail: 
                            #print("3")
                            dbdstart = exon[2]
                            dbdend = exon[2] - rbs.final + rbs.initial + subtract
                            if dbdstart > dbdend: dbdstart, dbdend = dbdend, dbdstart
                            dbd.add( GenomicRegion(chrom=exons[0][0], 
                                                   initial=dbdstart, final=dbdend, 
                                                   orientation=exons[0][3], 
                                                   name=str(cf)+"-"+str(rbs.final)+"_split2" ) )
                            dbdmap[str(rbs)] = dbd[-2].toString() + " & " + dbd[-1].toString() + " strand:-"
                            loop = False
                            break

                        elif rbs.initial > tail:
                            pass

                        cf += l
                        
                    loop = False
            else:

                while loop:
                    cf = 0
                    for exon in exons:
                        #print(exon)
                        l = exon[2] - exon[1]
                        tail = cf + l
                        #print("cf:   " + str(cf))
                        #print("tail: " + str(tail) )
                        if cf <= rbs.initial <=  tail:
                            dbdstart = exon[1] + rbs.initial - cf
                            
                            if rbs.final <= tail: 
                                #print("1")
                                dbdend = exon[1] + rbs.final -cf
                                dbd.add( GenomicRegion(chrom=exons[0][0], 
                                                       initial=dbdstart, final=dbdend, 
                                                       orientation=exons[0][3], 
                                                       name=str(rbs.initial)+"-"+str(rbs.final) ) )
                                dbdmap[str(rbs)] = dbd[-1].toString() + " strand:+"
                                loop = False
                                break
                            elif rbs.final > tail:

                                subtract = l + cf - rbs.initial
                                #print("2")
                                #print("Subtract: "+str(subtract))
                                dbd.add( GenomicRegion(chrom=exons[0][0], 
                                                       initial=dbdstart, final=exon[2], 
                                                       orientation=exons[0][3], 
                                                       name=str(rbs.initial)+"-"+str(rbs.initial+subtract)+"_split1" ) )

                        elif rbs.initial < cf and rbs.final <= tail: 
                            #print("3")
                            dbdstart = exon[1]
                            dbdend = exon[1] + rbs.final - rbs.initial - subtract
                            dbd.add( GenomicRegion(chrom=exons[0][0], 
                                                   initial=dbdstart, final=dbdend, 
                                                   orientation=exons[0][3], 
                                                   name=str(cf)+"-"+str(rbs.final)+"_split2" ) )
                            dbdmap[str(rbs)] = dbd[-2].toString() + " & " + dbd[-1].toString() + " strand:+"
                            loop = False
                            break

                        elif rbs.initial > tail:
                            pass

                        cf += l
                        
                    loop = False
        if not out_file:
            dbd.write_bed(filename=os.path.join(output, "DBD_"+rna_name+".bed"))
        else:
            # print(dbd)
            # print(dbd.sequences[0])
            dbd.write_bed(filename=output)
    # FASTA
    if fasta:
        #print(dbdmap)
        if not out_file:
            seq = pysam.Fastafile(os.path.join(output,"rna_temp.fa"))
            fasta_f = os.path.join(output, "DBD_"+rna_name+".fa")
        else:
            seq = pysam.Fastafile(os.path.join(temp,"rna_temp.fa"))
            fasta_f = output+".fa"

        with open(fasta_f, 'w') as fasta:
            for rbs in sig_region:
                print(">"+ rna_name +":"+str(rbs.initial)+"-"+str(rbs.final), file=fasta)
                s = seq.fetch(rbs.chrom, max(0, rbs.initial), rbs.final)
                for ss in [s[i:i + 80] for i in range(0, len(s), 80)]:
                    print(ss, file=fasta)

def connect_rna(rna, temp, rna_name):
    """Generate FASTA file merging all exons and return the number of exons and sequence length"""
    seq = ""
    exons = 0
    with open(rna) as f:
        for line in f:
            if line.startswith(">"):
                exons += 1
            else:
                line = line.strip()
                seq += line

    with open(os.path.join(temp,"rna_temp.fa"), "w") as r:
        print(">"+rna_name, file=r)
        for s in [seq[i:i + 80] for i in range(0, len(seq), 80)]:
            print(s, file=r)
    return [exons, len(seq)]

def get_dbss(input_BED,output_BED,rna_fasta,output_rbss,organism,l,e,c,fr,fm,of,mf,rm,temp):
    regions = GenomicRegionSet("Target")
    regions.read_bed(input_BED)
    regions.gene_association(organism=organism, show_dis=True)

    connect_rna(rna_fasta, temp=temp, rna_name="RNA")
    rnas = SequenceSet(name="rna", seq_type=SequenceType.RNA)
    rnas.read_fasta(os.path.join(temp,"rna_temp.fa"))
    rna_regions = get_rna_region_str(os.path.join(temp,rna_fasta))
    # print(rna_regions)
    genome = GenomeData(organism)
    genome_path = genome.get_genome()
    txp = find_triplex(rna_fasta=rna_fasta, dna_region=regions, 
                       temp=temp, organism=organism, remove_temp=False,
                       l=l, e=e, c=c, fr=fr, fm=fm, of=of, mf=mf, genome_path=genome_path,
                       prefix="targeted_region", dna_fine_posi=True)

    print("Total binding events:\t",str(len(txp)))
    txp.write_bed(output_BED)
    txp.write_txp(filename=output_BED.replace(".bed",".txp"))
    rbss = txp.get_rbs()
    dbd_regions(exons=rna_regions, sig_region=rbss, rna_name="rna", output=output_rbss, 
                out_file=True, temp=temp, fasta=False)
    # print(rbss.sequences)
    # print(len(rbss))
    # rbss.write_bed(output_rbss)

def get_rna_region_str(rna):
    """Getting the rna region from the information header with the pattern:
            REGION_chr3_51978050_51983935_-_
        or  chr3:51978050-51983935 -    """
    rna_regions = []
    with open(rna) as f:
        for line in f:
            if line[0] == ">":
                line = line.strip()
                # Loci
                if "REGION" in line:
                    l = line.split()
                    for i, e in enumerate(l):
                        if "REGION" in e:
                            e = e.split("_")
                            #print(e)
                            try:
                                r = [e[1], int(e[2]), int(e[3]), e[4]]
                            except:
                                r = [e[1], int(e[3]), int(e[4]), e[5]]
                
                elif "chr" in line:
                    try:
                        l = line.partition("chr")[2]
                        chrom = "chr" + l.partition(":")[0]
                        start = int(l.partition(":")[2].partition("-")[0])
                        end = int(l.partition(":")[2].partition("-")[2].split()[0])
                        sign = l.partition(":")[2].partition("-")[2].split()[1]
                        if sign == "+" or sign == "-" or sign == ".":
                            r = [chrom, start, end, sign]
                        elif "strand" in line:
                            s = l.partition("strand=")[2][0]
                            r = [chrom, start, end, s]
                        else:
                            print(line)
                    except:
                        break

                else:
                    r = []
                # Score
                if "score" in line:
                    ss = line.partition("score")[2].partition("=")[2]
                    score = float(ss.split()[0])
                    r.append(score)
                if r:
                    rna_regions.append(r)
    # if rna_regions:
    #     rna_regions.sort(key=lambda x: x[1])
    #     if rna_regions[0][3] == "-":
    #         rna_regions = rna_regions[::-1]

    return rna_regions


def no_binding_response(args, rna_regions, rna_name, organism, stat, expression):
    print("*** Find no DBD having DBS with cutoff = "+str(args.ccf))

    pro_path = os.path.join(os.path.dirname(args.o), "profile.txt")
    exp = os.path.basename(args.o)
    try:
        if args.de:
            tar = args.de
        else:
            tar = args.bed
    except:
        tar = args.bed
    stat["DBD_all"] = 0
    stat["DBSs_target_all"] = 0
    stat["DBSs_target_DBD_sig"] = 0

    save_profile(rna_regions=rna_regions, rna_name=rna_name,
                 organism=organism, output=args.o, bed=args.bed,
                 geneset=tar, stat=stat, topDBD=["-", 1], sig_DBD=[],
                 expression=expression)

    revise_index(root=os.path.dirname(os.path.dirname(args.o)))
    shutil.rmtree(args.o)
    sys.exit(1)

def write_stat(stat, filename):
    """Write the statistics into file"""
    order_stat = ["name", "genome", "exons", "seq_length",
                  "target_regions", "background_regions",
                  "DBD_all", "DBD_sig",
                  "DBSs_target_all", "DBSs_target_DBD_sig",
                  "DBSs_background_all", "DBSs_background_DBD_sig", "p_value"]
    with open(filename, "w") as f:
        for k in order_stat:
            try:
                print("\t".join([k,stat[k]]), file=f)
            except:
                continue

def integrate_stat(path):
    """Integrate all statistics within a directory"""
    base = os.path.basename(path)
    order_stat = ["name", "genome", "exons", "seq_length",
                  "target_regions", "background_regions",
                  "DBD_all", "DBD_sig",
                  "DBSs_target_all", "DBSs_target_DBD_sig",
                  "DBSs_background_all", "DBSs_background_DBD_sig", "p_value"]
    nested_dict = lambda: defaultdict(nested_dict)
    data = nested_dict()

    for item in os.listdir(path):
        pro = os.path.join(path, item, "stat.txt")
        if os.path.isfile(pro):
            with open(pro) as f:
                for line in f:
                    l = line.split()
                    data[item][l[0]] = l[1]
    with open(os.path.join(path,"statistics_"+base+".txt"), "w") as g:
        print("\t".join(order_stat), file=g)

        for item in data.keys():
            print("\t".join([data[item][o] for o in order_stat]), file=g)

def merge_DBD_regions(path):
    """Merge all available DBD regions in BED format. """

    for t in os.listdir(path):
        if os.path.isdir(os.path.join(path, t)):
            dbd_pool = GenomicRegionSet(t)
            for rna in os.listdir(os.path.join(path,t)):
                f = os.path.join(path, t, rna, "DBD_"+rna+".bed")
                if os.path.exists(f):
                    dbd = GenomicRegionSet(rna)
                    dbd.read_bed(f)
                    for r in dbd: r.name = rna+"_"+r.name
                    dbd_pool.combine(dbd)
            dbd_pool.write_bed(os.path.join(path, t, "DBD_"+t+".bed"))


def save_profile(rna_regions, rna_name, organism, output, bed,\
                 stat, topDBD, sig_DBD, expression, geneset=None):
    """Save statistics for comparison with other results"""

    pro_path = os.path.join(os.path.dirname(output), "profile.txt")
    exp = os.path.basename(output)
    # tag = os.path.basename(os.path.dirname(rnafile))
    if geneset:
        tar_reg = os.path.basename(geneset)
    else:
        tar_reg = os.path.basename(bed)
    # RNA name with region
    # if rna_regions:
    #     exon = str(len(rna_regions))
    # else:
    #     exon = "-"
    # rna = self.rna_name
    # RNA associated genes
    r_genes = rna_associated_gene(rna_regions=rna_regions, name=rna_name, organism=organism)

    newlines = []

    this_rna = [exp, rna_name, stat["exons"], stat["seq_length"],
                output.split("_")[-1], organism, tar_reg,
                value2str(float(stat["DBSs_target_all"])/int(stat["seq_length"])*1000),
                value2str(float(stat["DBSs_target_DBD_sig"])/int(stat["seq_length"])*1000),
                value2str(float(stat["DBD_all"])/int(stat["seq_length"])*1000), str(len(sig_DBD)),
                topDBD[0], value2str(topDBD[1]), r_genes, value2str(expression)]
    # try:
    if os.path.isfile(pro_path):
        with open(pro_path, 'r') as f:
            new_exp = True
            for line in f:
                line = line.strip()
                line = line.split("\t")
                if line[0] == exp:
                    newlines.append(this_rna)
                    new_exp = False
                elif line[0] == "Experiment":
                    continue
                else:
                    newlines.append(line)
            if new_exp:
                newlines.append(this_rna)
    else:
        newlines.append(this_rna)

    try: newlines.sort(key=lambda x: float(x[12]))
    except: pass
    newlines = [["Experiment", "RNA_names", "exon", "length",
                 "Tag", "Organism", "Target_region",
                 "Norm_DBS", "Norm_DBS_on_sig_DBD",
                 "Norm_DBD", "No_sig_DBDs", "Top_DBD",
                 "p-value", "closest_genes", "expression"]] + newlines

    with open(pro_path, 'w') as f:
        for lines in newlines:
            print("\t".join(lines), file=f)


# Import
import os
import sys

# Internal
from rgt.GenomicRegionSet import GenomicRegionSet
from rgt.Util import OverlapType

"""
Creates a bed file with MPBSs with (in green) and without (in red) evidence.
Also, the name of the instances will be Y for evidence, N for non evidence.

Authors: Eduardo G. Gusmao, Zhijian Li
"""
class Evidence:

    def __init__(self, peak_ext, mpbs_name, tfbs_summit_fname, mpbs_fname, output_location):
        self.peak_ext = peak_ext
        self.mpbs_name = mpbs_name
        self.tfbs_summit_fname = tfbs_summit_fname
        self.mpbs_fname = mpbs_fname
        self.output_location = output_location

    def create_file(self):
        # Expanding summits
        tfbs_summit_regions = GenomicRegionSet("TFBS Summit Regions")
        tfbs_summit_regions.read_bed(self.tfbs_summit_fname)

        for region in iter(tfbs_summit_regions):
            summit = int(region.data.split()[-1]) + region.initial
            region.initial = max(summit - (self.peak_ext / 2), 0)
            region.final = summit + (self.peak_ext / 2)

        # Calculating intersections
        mpbs_regions = GenomicRegionSet("MPBS Regions")
        mpbs_regions.read_bed(self.mpbs_fname)

        tfbs_summit_regions.sort()
        mpbs_regions.sort()

        with_overlap_regions = mpbs_regions.intersect(tfbs_summit_regions, mode=OverlapType.ORIGINAL)
        without_overlap_regions = mpbs_regions.subtract(tfbs_summit_regions, whole_region=True)
        tfbs_regions = GenomicRegionSet("TFBS Regions")

        for region in iter(with_overlap_regions):
            region.name = region.name.split(":")[0] + ":Y"
            tfbs_regions.add(region)

        for region in iter(without_overlap_regions):
            region.name = region.name.split(":")[0] + ":N"
            tfbs_regions.add(region)

        tfbs_regions.sort()

        tfbs_fname = os.path.join(self.output_location, "{}.bed".format(self.mpbs_name))
        tfbs_regions.write_bed(tfbs_fname)


"""
Util
===================
The Util classes contains many utilities needed by other classes such as the paths to input files.

"""

# Python
from __future__ import print_function
import os
import sys
import shutil
import ConfigParser
import traceback
from optparse import OptionParser,BadOptionError,AmbiguousOptionError


def npath(filename):
    """Returns a normalised, absolute version of the path, with expanded user directory."""
    return os.path.abspath(os.path.expanduser(filename))


class ConfigurationFile:
    """Represent the data path configuration file (data.config). It serves as a superclass to classes that will contain default variables (such as paths, parameters to tools, etc.) for a certain purpose (genomic data, motif data, etc.).

    *Variables:*

        - self.config -- Represents the configuration file.
        - self.data_dir -- Represents the root path to data files.

    """

    def __init__(self):
        
        # Reading config file directory
        root_dir = os.path.dirname(os.path.abspath(__file__))
        data_config_path_file_name = os.path.join(root_dir, 'data.config.path')
        data_config_path_file = open(data_config_path_file_name,"r")
        data_config_file_name = data_config_path_file.readline().strip()
        data_config_path_file.close()

        # Parsing config file
        self.config = ConfigParser.ConfigParser()
        self.config.read(data_config_file_name)

        # Overwriting config using user options
        self.config.read(data_config_file_name + ".user")

        # Reading data directory
        self.data_dir = os.path.split(data_config_file_name)[0]

class GenomeData(ConfigurationFile):
    """Represent genomic data. Inherits ConfigurationFile."""

    def __init__(self,organism):
        """Initializes GenomeData.

        *Keyword arguments:*

            - organism -- Organism alias.
        """
        ConfigurationFile.__init__(self)
        self.organism = organism
        self.genome = self.config.get(organism,'genome')
        self.chromosome_sizes = self.config.get(organism,'chromosome_sizes')
        # self.gene_regions = self.config.get(organism,'gene_regions')
        self.genes_gencode = self.config.get(organism, 'genes_Gencode')
        self.genes_refseq = self.config.get(organism, 'genes_RefSeq')
        self.annotation = self.config.get(organism,'annotation')
        self.annotation_dump_dir = os.path.dirname(self.annotation)
        self.gene_alias = self.config.get(organism,'gene_alias')

    def get_organism(self):
        """Returns the current organism."""
        return self.organism

    def get_genome(self):
        """Returns the current path to the genome fasta file."""
        return self.genome

    def get_chromosome_sizes(self):
        """Returns the current path to the chromosome sizes text file."""
        return self.chromosome_sizes   
    
    def get_gene_regions(self):
        """Returns the current path to the gene_regions BED file."""
        return self.genes_gencode

    def get_genes_gencode(self):
        """Returns the current path to the gene_regions BED file."""
        return self.genes_gencode

    def get_genes_refseq(self):
        """Returns the current path to the gene_regions BED file."""
        return self.genes_refseq

    def get_annotation(self):
        """
        Returns the current path to the gencode annotation gtf file.
        """
        return self.annotation

    def get_annotation_dump_dir(self):
        """Returns the current path to the gencode annotation gtf file."""
        return self.annotation_dump_dir

    def get_gene_alias(self):
        """Returns the current path to the gene alias txt file."""
        return self.gene_alias



class MotifData(ConfigurationFile):
    """Represent motif (PWM) data. Inherits ConfigurationFile."""

    def __init__(self):
        ConfigurationFile.__init__(self)
        self.repositories_list = self.config.get('MotifData','repositories').split(",")
        self.pwm_list = []
        self.logo_list = []
        self.mtf_list = []
        self.fpr_list = []
        for current_repository in self.repositories_list:
            self.pwm_list.append(os.path.join(self.data_dir,self.config.get('MotifData','pwm_dataset'),current_repository))
            self.logo_list.append(os.path.join(self.data_dir,self.config.get('MotifData','logo_dataset'),current_repository))
            self.mtf_list.append(os.path.join(self.data_dir,self.config.get('MotifData','pwm_dataset'),current_repository+".mtf"))
            self.fpr_list.append(os.path.join(self.data_dir,self.config.get('MotifData','pwm_dataset'),current_repository+".fpr"))

    def get_repositories_list(self):
        """Returns the current repository list."""
        return self.repositories_list

    def get_pwm_path(self, current_repository):
        """Returns the path to a specific motif repository."""
        return os.path.join(self.data_dir,self.config.get('MotifData','pwm_dataset'),current_repository)

    def get_pwm_list(self):
        """Returns the list of current paths to the PWM repositories."""
        return self.pwm_list

    def get_logo_file(self, current_repository):
        """Returns the path to a specific logo repository.

        *Keyword arguments:*

            - current_repository -- Motif repository.
        """
        return os.path.join(self.data_dir,self.config.get('MotifData','logo_dataset'),current_repository)

    def get_logo_list(self):
        """Returns the list of current paths to the logo images of PWMs in the given repositories."""
        return self.logo_list

    def get_mtf_path(self, current_repository):
        """Returns the path to a specific mtf file.

        *Keyword arguments:*

            - current_repository -- Motif repository.
        """
        return os.path.join(self.data_dir,self.config.get('MotifData','pwm_dataset'),current_repository+".mtf")

    def get_mtf_list(self):
        """Returns the list of current paths to the mtf files."""
        return self.mtf_list

    def get_fpr_list(self):
        """Returns the list of current paths to the fpr files."""
        return self.fpr_list

class HmmData(ConfigurationFile):
    """Represent HMM data. Inherits ConfigurationFile."""

    def __init__(self):
        ConfigurationFile.__init__(self)
        self.default_hmm_dnase = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_dnase'))
        self.default_hmm_dnase_bc = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_dnase_bc'))
        self.default_hmm_atac = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_atac'))
        self.default_hmm_atac_bc = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_atac_bc'))
        self.default_hmm_histone = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_histone'))
        self.default_hmm_dnase_histone = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_dnase_histone'))
        self.default_hmm_dnase_histone_bc = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_dnase_histone_bc'))
        self.default_hmm_atac_histone = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_atac_histone'))
        self.default_hmm_atac_histone_bc = os.path.join(self.data_dir,self.config.get('HmmData','default_hmm_atac_histone_bc'))
        self.default_bias_table_F_SH = os.path.join(self.data_dir,self.config.get('HmmData','default_bias_table_F_SH'))
        self.default_bias_table_R_SH = os.path.join(self.data_dir,self.config.get('HmmData','default_bias_table_R_SH'))
        self.default_bias_table_F_DH = os.path.join(self.data_dir,self.config.get('HmmData','default_bias_table_F_DH'))
        self.default_bias_table_R_DH = os.path.join(self.data_dir,self.config.get('HmmData','default_bias_table_R_DH'))
        self.default_bias_table_F_ATAC = os.path.join(self.data_dir,self.config.get('HmmData','default_bias_table_F_ATAC'))
        self.default_bias_table_R_ATAC = os.path.join(self.data_dir,self.config.get('HmmData','default_bias_table_R_ATAC'))

    def get_default_hmm_dnase(self):
        """Returns the current default DNase only hmm."""
        return self.default_hmm_dnase

    def get_default_hmm_dnase_bc(self):
        """Returns the current default DNase only hmm."""
        return self.default_hmm_dnase_bc

    def get_default_hmm_atac(self):
        """Returns the current default atac only hmm."""
        return self.default_hmm_atac

    def get_default_hmm_atac_bc(self):
        """Returns the current default atac only hmm."""
        return self.default_hmm_atac_bc

    def get_default_hmm_histone(self):
        """Returns the current default Histone only hmm."""
        return self.default_hmm_histone

    def get_default_hmm_dnase_histone(self):
        """Returns the current default DNase+histone hmm."""
        return self.default_hmm_dnase_histone

    def get_default_hmm_dnase_histone_bc(self):
        """Returns the current default DNase+histone hmm."""
        return self.default_hmm_dnase_histone_bc

    def get_default_hmm_atac_histone(self):
        """Returns the current default atac+histone hmm."""
        return self.default_hmm_atac_histone

    def get_default_hmm_atac_histone_bc(self):
        """Returns the current default atac+histone hmm."""
        return self.default_hmm_atac_histone_bc

    def get_default_bias_table_F_SH(self):
        """Returns the DNase-seq single hit default bias table for the forward strand."""
        return self.default_bias_table_F_SH

    def get_default_bias_table_R_SH(self):
        """Returns the DNase-seq single hit default bias table for the reverse strand."""
        return self.default_bias_table_R_SH

    def get_default_bias_table_F_DH(self):
        """Returns the DNase-seq double hit default bias table for the forward strand."""
        return self.default_bias_table_F_DH

    def get_default_bias_table_R_DH(self):
        """Returns the DNase-seq double hit default bias table for the reverse strand."""
        return self.default_bias_table_R_DH

    def get_default_bias_table_F_ATAC(self):
        """Returns the ATAC-seq default bias table for the forward strand."""
        return self.default_bias_table_F_ATAC

    def get_default_bias_table_R_ATAC(self):
        """Returns the ATAC-seq default bias table for the reverse strand."""
        return self.default_bias_table_R_ATAC

class ImageData(ConfigurationFile):
    """Represent image data. Inherits ConfigurationFile."""

    def __init__(self):
        ConfigurationFile.__init__(self)
        self.rgt_logo = os.path.join(self.data_dir,"fig","rgt_logo.gif")
        self.css_file = os.path.join(self.data_dir,"fig","style.css")
        self.default_motif_logo = os.path.join(self.data_dir,"fig","default_motif_logo.png")
        self.tablesorter = os.path.join(self.data_dir,"fig","jquery.tablesorter.min.js")
        self.jquery = os.path.join(self.data_dir,"fig","jquery-1.11.1.js")
        self.jquery_metadata = os.path.join(self.data_dir,"fig","jquery.metadata.js")
        self.tdf_logo = os.path.join(self.data_dir,"fig","tdf_logo.png")
        self.viz_logo = os.path.join(self.data_dir,"fig","viz_logo.png")


    def get_rgt_logo(self):
        """Returns the rgt logo image file location."""
        return self.rgt_logo

    def get_css_file(self):
        """Returns the css file location."""
        return self.css_file

    def get_default_motif_logo(self):
        """Returns the default motif logo file location."""
        return self.default_motif_logo
    
    def get_sorttable_file(self):
        """Returns the default sorttable code location."""
        return self.default_motif_logo
    
    def get_jquery(self):
        """Returns the jquery code location."""
        return self.jquery

    def get_tablesorter(self):
        """Returns the table sorter code location."""
        return self.tablesorter
    
    def get_jquery_metadata(self):
        """Returns the jquery metadata location."""
        return self.jquery_metadata
        
    def get_tdf_logo(self):
        """Returns the default TDF logo."""
        return self.tdf_logo

    def get_viz_logo(self):
        """Returns the default RGT viz logo."""
        return self.viz_logo

class Library_path(ConfigurationFile):
    """Represent the path to triplexator. Inherits ConfigurationFile."""

    def __init__(self):
        ConfigurationFile.__init__(self)
        self.path_triplexator = self.config.get("Library",'path_triplexator')

        self.path_c_rgt = self.config.get("Library", 'path_c_rgt')

    def get_triplexator(self):
        return self.path_triplexator
    def get_c_rgt(self):
        return self.path_c_rgt


class OverlapType:
    """Class of overlap type constants.

    *Constants:*

        - OVERLAP -- Return new GenomicRegionSet including only the overlapping regions.
        - ORIGINAL -- Return the regions of original GenomicRegionSet which have any intersections.
        - COMP_INCL -- Return region(s) of the GenomicRegionSet which are 'completely' included.
    """

    OVERLAP = 0 
    ORIGINAL = 1
    COMP_INCL = 2

class SequenceType:
    """Class of sequence type
    
    *Constants:*
    
        - DNA
        - RNA
    """
    DNA = 0
    RNA = 1

class HelpfulOptionParser(OptionParser):
    """An OptionParser that prints full help on errors. Inherits OptionParser."""
    def error(self, msg):
        self.print_help(sys.stderr)
        self.exit(2, "\n%s: error: %s\n" % (self.get_prog_name(), msg))

class PassThroughOptionParser(HelpfulOptionParser):
    """When unknown arguments are encountered, bundle with largs and try again, until rargs is depleted. sys.exit(status) will still be called if a known argument is passed incorrectly (e.g. missing arguments or bad argument types, etc.). Inherits HelpfulOptionParser."""
    def _process_args(self, largs, rargs, values):
        while rargs:
            try:
                HelpfulOptionParser._process_args(self,largs,rargs,values)
            except (BadOptionError,AmbiguousOptionError), e:
                pass
                #largs.append(e.opt_str)

class ErrorHandler():
    """Handles errors in a standardized way.

        *Error Dictionary Standard:*

            Each entry consists of a key+list in the form X:[Y,Z,W] where:

            - X -- The key representing the internal error name.
            - Y -- Error number.
            - Z -- Exit status.
            - W -- Error message to be print.

        *Warning Dictionary Standard:*

            Each entry consists of a key+list in the form X:[Y,Z] where:

            - X -- The key representing the internal warning name.
            - Y -- Warning number.
            - Z -- Warning message to be print.
    """

    def __init__(self):

        self.program_name = os.path.basename(sys.argv[0])

        self.error_dictionary = {
            "DEFAULT_ERROR": [0,0,"Undefined error. Program terminated with exit status 0."],
            "MOTIF_ANALYSIS_OPTION_ERROR": [1,0,"You must define one specific analysis. Run '"+self.program_name+" -h' for help."],

            "FP_WRONG_ARGUMENT": [2,0,"You must provide at least one and no more than one experimental matrix as input argument."],
            "FP_WRONG_EXPMAT": [3,0,"The experimental matrix could not be loaded. Check if it is correctly formatted and that your python version is >= 2.7."],
            "FP_ONE_REGION": [4,0,"You must provide one 'regions' bed file in the experiment matrix."],
            "FP_NO_DNASE": [5,0,"You must provide one 'reads' file termed DNASE or ATAC in the experiment matrix."],
            "FP_NO_HISTONE": [6,0,"You must provide at least one 'reads' file not termed DNASE or ATAC (histone modification) in the experiment matrix."],
            "FP_NB_HMMS": [7,0,"You must provide one HMM file or X HMM files where X is the number of histone tracks detected in the experiment matrix."],
            "FP_HMM_FILES": [8,0,"Your HMM file could not be read. If you did not provide any HMM file, you may not have installed scikit correctly."],
            "FP_BB_CREATION": [9,0,"Big Bed file (.bb) could not be created. Check if you have the bedToBigBed script in $PATH."],
            "MM_OUT_FOLDER_CREATION": [10,0,"Could not create output folder."],
            "MM_NO_ARGUMENT": [11,0,"Could not read the arguments. Make sure you provided an experimental matrix."],
            "MM_WRONG_EXPMAT": [12,0,"The experimental matrix could not be loaded. Check if it is correctly formatted and that your python version is >= 2.7."],
            "MM_WRONG_RANDPROP": [13,0,"Proportion of random regions is too low (<= 0)."],
            "MM_LOW_NPROC": [14,0,"Number of processors is too low (<= 0)."],
            "ME_OUT_FOLDER_CREATION": [15,0,"Could not create output folder."],
            "ME_FEW_ARG": [16,0,"There are too few arguments. Please use -h option in order to verify command usage."],
            "ME_WRONG_EXPMAT": [17,0,"The experimental matrix could not be loaded. Check if it is correctly formatted and that your python version is >= 2.7."],
            "ME_MATCH_NOTFOUND": [18,0,"Motif matching file was not found. Are you sure you performed --matching before --enrichment?"],
            "ME_BAD_MATCH": [19,0,"Motif matching file is incorrect. Please perform --matching again."],
            "ME_LOW_NPROC": [20,0,"Number of processor is too low (<= 0)."],
            "ME_RAND_NOTFOUND": [21,0,"Random regions not found. Are you sure --matching was correctly performed?"],
            "ME_BAD_RAND": [22,0,"Could not read random regions."],
            "ME_RAND_NOT_BED_BB": [23,0,"Random regions are not in bed or bigbed format."],
            "MM_PSEUDOCOUNT_0": [24,0,"There were errors involved in the creation of Position Weight Matrices. Some distributions of numpy  and/or scipy does not allow for pseudocounts == 0. Please increase the pseudocount (or use default value of 0.1) and try again."],
            "MM_MOTIFS_NOTFOUND": [25,0,"A motif file was provided but it could not be loaded."],
            "XXXXXXX": [26,0,"Xxxxxx"]
        }
        self.error_number = 0
        self.exit_status = 1
        self.error_message = 2

        self.warning_dictionary = {
            "DEFAULT_WARNING": [0,"Undefined warning."],
            "FP_ONE_REGION": [1,"There are more than one 'regions' file in the experiment matrix. Only the first will be used."],
            "FP_MANY_DNASE": [2,"There are more than one DNASE or ATAC 'reads' file. Only the first one will be used."],
            "FP_MANY_HISTONE": [3,"It is recomended that no more than three histone modifications should be used."],
            "FP_DNASE_PROC": [4,"The DNase  or ATAC file could not be processed."],
            "FP_HISTONE_PROC": [5,"The Histone file could not be processed."],
            "FP_SEQ_FORMAT": [6,"The DNase/ATAC+Histone sequence could not be formatted to be input for scikit."],
            "FP_HMM_APPLIC": [7,"The scikit HMM encountered errors when applied."],
            "MM_MANY_ARG": [8,"There are more than one arguments, only the first experimental matrix will be used."],
            "ME_MANY_ARG": [9,"There are more than two arguments, only the two first arguments will be used."],
            "ME_MANY_GENESETS": [10,"There are more than one geneset associated to a group, only the first one will be used."],
            "ME_FEW_GENESETS": [11,"There seems to be a geneset column, but no geneset was found for some entries."],
            "XXXXXXX": [12,"Xxxxxx"]
        }
        self.warning_number = 0
        self.warning_message = 1

    def throw_error(self, error_type, add_msg = ""):
        """Throws the specified error type. If the error type does not exist, throws a default error message and exits.

        *Keyword arguments:*

            - error_type -- Error type.
            - add_msg -- Message to add to the error.
        """

        # Fetching error type
        try:
            error_number = self.error_dictionary[error_type][self.error_number]
            exit_status = self.error_dictionary[error_type][self.exit_status]
            error_message = self.error_dictionary[error_type][self.error_message]
        except KeyError, IndexError:
            error_number = self.error_dictionary["DEFAULT_ERROR"][self.error_number]
            exit_status = self.error_dictionary["DEFAULT_ERROR"][self.exit_status]
            error_message = self.error_dictionary["DEFAULT_ERROR"][self.error_message]

        # Handling error
        complete_error_message = ("--------------------------------------------------\n"
                                  "Error Number: "+str(error_number)+".\n"
                                  "Program: "+self.program_name+".\n"
                                  "Report: "+error_message+" "+add_msg+"\n"
                                  "Behaviour: The program will quit with exit status "+str(exit_status)+".\n"
                                  "--------------------------------------------------")
        print(complete_error_message, file=sys.stderr)
        traceback.print_exc()
        sys.exit(exit_status)

    def throw_warning(self, warning_type, add_msg = ""):
        """Throws the specified warning type. If the warning type does not exist, throws a default warning message and exits.

        *Keyword arguments:*

            - warning_type -- Warning type.
            - add_msg -- Message to add to the error.
        """

        # Fetching warning type
        try:
            warning_number = self.warning_dictionary[warning_type][self.warning_number]
            warning_message = self.warning_dictionary[warning_type][self.warning_message]
        except KeyError, IndexError:
            warning_number = self.warning_dictionary["DEFAULT_WARNING"][self.warning_number]
            warning_message = self.warning_dictionary["DEFAULT_WARNING"][self.warning_message]

        # Handling warning
        complete_warning_message = ("--------------------------------------------------\n"
                                    "Warning Number: "+str(warning_number)+".\n"
                                    "Program: "+self.program_name+".\n"
                                    "Report: "+warning_message+" "+add_msg+"\n"
                                    "--------------------------------------------------")
        print(complete_warning_message, file=sys.stderr)

class Html:
    """Represent an HTML file.

    *Keyword arguments:*

        - name -- Name of the HTML document.
        - links_dict -- Dictionary with the upper links.
        - fig_dir -- Figure directory (default = None).
        - fig_rpath -- Relative figure path (default = '../fig').
        - cluster_path_fix -- deprecated.
        - RGT_header -- Whether to print RGT header (default = True).
        - other_logo -- Other tool logos (default = None).
        - homepage -- Homepage link (default = None).

    .. warning:: cluster_path_fix is going to be deprecated soon. Do not use it.
    """

    def __init__(self, name, links_dict, fig_dir=None, fig_rpath="../fig", cluster_path_fix="", 
                 RGT_header=True, other_logo=None, homepage=None):

        # Variable initializations
        self.name = name
        self.links_dict = links_dict
        self.cluster_path_fix = cluster_path_fix
        self.document = []
        self.image_data = ImageData()
        self.other_logo = other_logo
        self.homepage = homepage
        
        # Initialize document
        if fig_dir:
            self.copy_relevent_files(fig_dir)
            self.create_header(relative_dir=fig_rpath, RGT_name=RGT_header, other_logo=other_logo)
        else:
            self.create_header(relative_dir=fig_rpath, RGT_name=RGT_header, other_logo=other_logo)
        
        self.add_links()

    def copy_relevent_files(self, target_dir):
        """Copies relevant files to relative paths.

        *Keyword arguments:*

            - target_dir -- Target directory to copy files.
        """

        try:
            os.stat(target_dir)
        except:
            os.mkdir(target_dir)
        shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_rgt_logo(), dst=os.path.join(target_dir,"rgt_logo.gif"))
        shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_css_file(), dst=os.path.join(target_dir,"style.css"))
        #shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_jquery(), dst=os.path.join(target_dir,"jquery-1.11.1.js"))
        shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_tablesorter(), dst=os.path.join(target_dir,"jquery.tablesorter.min.js"))
        #shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_jquery_metadata(), dst=os.path.join(target_dir,"jquery.metadata.js"))
        #shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_tablesorter(), dst=os.path.join(target_dir,"jquery.metadata.js"))
        if self.other_logo:
            if self.other_logo == "TDF":
                shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_tdf_logo(), dst=os.path.join(target_dir,"tdf_logo.png"))
            if self.other_logo == "viz":
                shutil.copyfile(src=self.cluster_path_fix+self.image_data.get_viz_logo(), dst=os.path.join(target_dir,"viz_logo.png"))
          
    def create_header(self, relative_dir=None, RGT_name=True, other_logo=None):
        """Creates default document header.

        *Keyword arguments:*

            - relative_dir -- Define the directory to store CSS file and RGT logo so that the html code can read from it (default = None).
            - RGT_name -- Whether to print RGT name (default = True).
            - other_logo -- Other tool logos (default = None)
        """

        self.document.append('<script src="http://ajax.googleapis.com/ajax/libs/jquery/1.11.2/jquery.min.js"></script>') 
            
        if relative_dir:
            #self.document.append('<script type="text/javascript" src="'+relative_dir+'/jquery-1.11.1.js"></script>')
            self.document.append('<script type="text/javascript" src="'+relative_dir+'/jquery.tablesorter.min.js"></script>')
            #self.document.append('<script type="text/javascript" src="'+relative_dir+'/jquery.metadata.js"></script>')
        else:
            #self.document.append('<script type="text/javascript" src="'+self.cluster_path_fix+self.image_data.get_jquery()+'"></script>')
            self.document.append('<script type="text/javascript" src="'+self.cluster_path_fix+self.image_data.get_tablesorter()+'"></script>')
            #self.document.append('<script type="text/javascript" src="'+self.cluster_path_fix+self.image_data.get_jquery_metadata()+'"></script>')
        
        self.document.append("<html>")
        self.document.append("<head><meta http-equiv=\"Content-Type\" content=\"text/html\"><title>RGT "+self.name+"</title>")

        self.document.append("<style type=\"text/css\">")
        self.document.append("<!--")
        
        if relative_dir:
            self.document.append("@import url(\""+relative_dir+"/style.css\");")
        else:
            self.document.append("@import url(\""+self.cluster_path_fix+self.image_data.get_css_file()+"\");")
        
        self.document.append("-->")
        self.document.append("</style></head>")
        self.document.append("<body topmargin=\"0\" leftmargin=\"0\" rightmargin=\"0\" bottommargin=\"0\" marginheight=\"0\" marginwidth=\"0\" bgcolor=\"#FFFFFF\">")        
        
        self.document.append("<h3 style=\"background-color:white; border-top:3px solid gray; border-bottom:3px solid gray;\">")
        self.document.append("<table border=\"0\" width=\"100%\" cellspacing=\"0\" cellpadding=\"0\">")
        self.document.append("  <tr>")

        # Logo
        if relative_dir:            
            self.document.append("    <td width=\"5%\">")
            if self.homepage: self.document.append("<a href=\""+self.homepage+"\">")
            if other_logo == "TDF":
                self.document.append("    <img border=\"0\" src=\""+relative_dir+"/tdf_logo.png"+"\" width=\"130\" height=\"100\">")
            elif other_logo == "viz":
                self.document.append("    <img border=\"0\" src=\""+relative_dir+"/viz_logo.png"+"\" width=\"130\" height=\"100\">")
            else:
                self.document.append("    <img border=\"0\" src=\""+relative_dir+"/rgt_logo.gif\" width=\"130\" height=\"100\">")
            if self.homepage: self.document.append("</a>")
            self.document.append("    </td>")
            
        else:
            self.document.append("    <td width=\"5%\"><img border=\"0\" src=\""+self.cluster_path_fix+self.image_data.get_rgt_logo()+"\" width=\"130\" height=\"100\"></td>")

        # Gap
        self.document.append("     <td width=\"5%\"></td>")
        # Title
        if RGT_name:
            self.document.append("    <td width=\"90%\"><p align=\"left\"><font color=\"black\" size=\"5\">Regulatory Genomics Toolbox - "+self.name+"</font></td>")
        else:
            self.document.append("    <td width=\"90%\"><p align=\"left\"><font color=\"black\" size=\"5\">"+self.name+"</font></td>")
        
        self.document.append("  </tr>")
        self.document.append("</table>")
        self.document.append("</h3>")

    def add_links(self):
        """Adds all the links."""
        for k in self.links_dict.keys():

            self.document.append("<a class=\"pure-button\" href=\""+\
                                 os.path.join(self.cluster_path_fix,self.links_dict[k])+\
                                 "\">"+\
                                 "<font size='1'>"+k+"</font>"+"</a>")

        #self.document.append("<table border=\"0\" width=\"100%\" cellspacing=\"0\" cellpadding=\"0\">")
        #self.document.append("  <tr>")
        #self.document.append("    <td width=\"100%\"><font color=\"black\" face=\"Arial\" size=\"4\"><b>&nbsp;&nbsp;")
        #link_str = "    "+" &nbsp;&nbsp; |&nbsp;&nbsp; ".join(["<a href=\""+os.path.join(self.cluster_path_fix,self.links_dict[k])+"\">"+k+"</a>" for k in self.links_dict.keys()])
        #self.document.append(link_str)
        #self.document.append("    </b></font></td>")
        #self.document.append("  </tr>")
        #self.document.append("</table>")

    def create_footer(self):
        """Adds footer."""
        self.document.append("<br><br>")
        self.document.append("<p align=\"center\"><font face=\"Arial\" color=\"#000000\" size=\"2\">")
        self.document.append("For more details please visit the <a href=\"http://www.regulatory-genomics.org/\"> RGT Website </a>")
        self.document.append("</font></p>")
        self.document.append("<h3 style=\"background-color:white; border-top:3px solid gray;\"></h3>")
        self.document.append("</body>")
        self.document.append("</html>")

    def add_heading(self, heading, align = 50, color = "black", face = "Arial", size = 5, bold = True, idtag=None):
        """Creates a heading.
        
        *Keyword arguments:*

            - heading -- The heading title.
            - align -- Alignment of the heading. Can be either an integer (interpreted as left margin) or string (interpreted as HTML positional argument (default = 50).
            - color -- Color of the heading (default = "black").
            - face -- Font of the heading (default = "Arial").
            - size -- Size of the heading (HTML units [1,7]) (default = 5).
            - bold -- Whether the heading is bold (default = True).
            - idtag -- Add ID tag in the heading element (default = None).
        """

        if idtag:
            idstr = ' id="'+idtag+'"'
        else:
            idstr = ""

        # Creating header
        content_str = ""
        if(isinstance(align,int)): content_str += "<p style=\"margin-left: "+str(align)+"\""+idstr+">"
        elif(isinstance(align,str)): content_str += "<p align=\""+align+"\""+idstr+">"
        else: pass # ERROR
        content_str += "<font color=\""+color+"\" face=\""+face+"\" size=\""+str(size)+"\""+idstr+">"
        if(bold): content_str += "<b>"
        self.document.append(content_str)

        # Printing heading name
        self.document.append(heading)

        # Creating footing
        end_str = ""
        if(bold): end_str += "</b>"
        end_str += "</font></p>"
        self.document.append(end_str)

    def add_zebra_table(self, header_list, col_size_list, type_list, data_table, align = 50, 
                        cell_align = 'center', auto_width=False, colorcode=None, header_titles=None,
                        border_list=None, sortable=False):
        """Creates a zebra table.

        *Keyword arguments:*

            - header_list -- A list with the table headers in correct order.
            - col_size_list -- A list with the column sizes (integers).
            - type_list -- A string in which each character represents the type of each row.
                  - s = string (regular word or number)
                  - i = image
                  - l = link 
            - data_table -- A table containing the data to be input according to each data type defined.
                  - s = string
                  - i = tuple containing: ("file name", width) width = an integer
                  - l = tuple containing: ("Name","Link")
            - align -- Alignment of the heading. Can be either an integer (interpreted as left margin) or string (interpreted as HTML positional argument) (default = 50).
            - cell_align -- Alignment of each cell in the table (default = center).
            - auto_width -- Adjust the column width by the content automatically regardless of defined col size (default = False).
            - colorcode -- Color code (default = None)
            - header_titles -- Given a list corresponding to the header_list, which defines all the explanation in hint windows (default = None).
            - border_list -- Table borders (default = None).
            - sortable -- Whether it is a sortable table (default = False).
        """
        #if header_notes: self.document.append("<style> .ami div {display:none} .ami:hover div {display:block} </style>")
        
        if not border_list:
            border_list = [""] * len(data_table[0])
        if auto_width: auto= " table-layout: auto"
        else: auto=""


        # Starting table
        type_list = type_list.lower()
        if(isinstance(align,int)): self.document.append("<p style=\"margin-left: "+str(align)+"\">")
        elif(isinstance(align,str)): self.document.append("<p align=\""+align+"\">") 
        else: pass # TODO ERROR
        
        # Table header
        #self.document.append("<table id=\"myTable\" class=\"tablesorter\">")
        
        if sortable: 
            sortableclass=" class=\"tablesorter\""
            tableid="sortable"
        else:
            sortableclass=""
            tableid="hor-zebra"

        self.document.append("<table id=\""+tableid+"\""+sortableclass+auto+">")

        if colorcode:
            for line in colorcode:
                self.document.append(line)

        #############################
        ##### Header ################
        self.document.append("  <thead>")
        if (isinstance(header_list[0], list)):
        # For headers more than one row
            for r, row_list in enumerate(header_list):
                self.document.append("    <tr>")
                header_str = []

                merge_num = [1] * len(row_list)
                for i, value in enumerate(row_list):
                    if value == None:
                        merge_num[last_true] += 1
                        merge_num[i] -= 1
                    else:
                        last_true = i

                for i in range(0,len(row_list)):
                    if merge_num[i] > 1:
                        if header_titles:
                            header_str.append("<th scope=\"col\" width=\""+str(col_size_list[i])+
                                              "\" align=\""+'center'+"\""+" colspan=\""+str(merge_num[i])+"\" "+
                                              "title=\""+header_titles[r][i]+"\""+border_list[i+merge_num[i]-1]+auto+" >"+
                                              row_list[i]+"</th>")
                        else:
                            header_str.append("<th scope=\"col\" width=\""+str(col_size_list[i])+
                                              "\" align=\""+'center'+"\""+" colspan=\""+str(merge_num[i])+"\""+
                                              border_list[i+merge_num[i]-1]+auto+">"+row_list[i]+"</th>")
                        
                    elif merge_num[i] == 0:
                        continue
                    else:
                        if header_titles:
                            header_str.append("<th scope=\"col\" width=\""+str(col_size_list[i])+
                                              "\" align=\""+cell_align+"\" "+
                                              "title=\""+header_titles[r][i]+"\""+border_list[i]+auto+">"+
                                              row_list[i]+"</th>")
                        else:
                            header_str.append("<th scope=\"col\" width=\""+str(col_size_list[i])+
                                              "\" align=\""+cell_align+"\""+border_list[i]+auto+">"+
                                              row_list[i]+"</th>")

                header_str = "    "+"\n    ".join(header_str)
                self.document.append(header_str)
                self.document.append("    </tr>")

        else:
            self.document.append("    <tr>")
            header_str = []
            for i in range(0,len(header_list)):
                if header_titles:
                    header_str.append("<th scope=\"col\" width=\""+str(col_size_list[i])+"\" align=\""+cell_align+"\" "+
                                      "title=\""+header_titles[i]+"\" >"+header_list[i]+"</th>")
                else:
                    header_str.append("<th scope=\"col\" width=\""+str(col_size_list[i])+"\" align=\""+cell_align+"\">"+
                                      header_list[i]+"</th>")
                
            header_str = "    "+"\n    ".join(header_str)
            self.document.append(header_str)
            self.document.append("    </tr>")
        self.document.append("  </thead>")

        
        #############################
        ##### Table body ############
        self.document.append("  <tbody>")
        for i in range(0,len(data_table)):

            # Row type
            if(i%2==0) and not sortable: self.document.append("    <tr class=\"odd\">")
            else: self.document.append("    <tr>")

            # Body data
            for j in range(0,len(data_table[i])):
                if(type_list[j] == "s"):
                    self.document.append("      <td align=\""+cell_align+"\" "+border_list[j]+">"+data_table[i][j]+"</td>")
                elif(type_list[j] == "i"): 
                    self.document.append("      <td align=\""+cell_align+"\"><img src=\""+self.cluster_path_fix+
                                         data_table[i][j][0]+"\" width="+str(data_table[i][j][1])+" ></td>")
                elif(type_list[j] == "l"):
                    self.document.append("      <td align=\""+cell_align+"\"><a href=\""+data_table[i][j][1]+"\">"+
                                         data_table[i][j][0]+"</a></td>")
                else: pass # TODO ERROR

            # Row ending
            self.document.append("    </tr>")
        
        # Finishing table
        self.document.append("</tbody></table></p>")

    def add_fixed_rank_sortable(self):
        """Add jquery for fixing the first column of the sortable table"""
        scripts = ["<script>",
                   "// add custom numbering widget",
                   "$.tablesorter.addWidget({",
                   "    id: 'numbering',",
                   "    format: function(table) {",
                   "        var c = table.config;",
                   "        $('tr:visible', table.tBodies[0]).each(function(i) {",
                   "            $(this).find('td').eq(0).text(i + 1);",
                   "        });",
                   "    }",
                   "});",
                   "$('.tablesorter').tablesorter({",
                   "    // prevent first column from being sortable",
                   "    headers: {",
                   "        0: { sorter: false }",
                   "    },",
                   "    // apply custom widget",
                   "    widgets: ['numbering']",
                   "});",
                   "</script>"]
        for s in scripts:
            self.document.append(s)

    def add_figure(self, figure_path, notes=None, align=50, color="black", face="Arial", size=3, 
                   bold=False, width="800", more_images=None):
        """Add a figure with notes underneath.
        
        *Keyword arguments:*

            - figure_path -- The path to the figure.
            - notes -- A list of strings for further explanation
            - align -- Alignment of the heading. Can be either an integer (interpreted as left margin) or string (interpreted as HTML positional argument) (default = 50).
            - color -- Color (default = 'black').
            - face -- Font (default = 'Arial').
            - size -- Size (default = 3).
            - bold -- Whether it is bold (default = False).
            - width -- Width (default = 800).
            - more_images -- Add more images (default = None).
        """        
        if(isinstance(align,int)): img_str = "<p style=\"margin-left: "+str(align)+"\">"
        elif(isinstance(align,str)): img_str = "<p align=\""+str(align)+"\">"
        else: pass # TODO ERROR
        
        img_str += '<a href="'+figure_path+'"><img src="'+ figure_path +'" width='+width+'></a>'
        
        if more_images:
            for im in more_images:
                img_str += '<a href="'+im+'"><img src="'+ im +'" width='+width+'></a>'
        
        img_str += '</p>'

        self.document.append(img_str)
        if notes:
            if(isinstance(align,int)): 
                note_str = "<p style=\"margin-left: "+str(align)+"\"><font color=\""+color+"\" face=\""+face+"\" size=\""+str(size)+"\ align=\""+ str(align) + "\">"
            elif(isinstance(align,str)):
                note_str = "<p align=\""+str(align)+"\"><font color=\""+color+"\" face=\""+face+"\" size=\""+str(size)+"\ align=\""+ str(align) + "\">"            
            else: pass # TODO ERROR
            
            if(bold): note_str += "<b>"
            for line in notes:
                note_str += line + "<br>" 
            if(bold): note_str += "</b>"
            note_str += "</font></p>"
            self.document.append(note_str)

    def add_free_content(self, content_list):
        """Adds free HTML to the document.

        *Keyword arguments:*

            - content_list -- List of strings. Each string is interpreted as a line in the HTML document.
        """
        for e in content_list: self.document.append(e)

    def add_list(self, list_of_items, ordered=False):
        """Add a list to the document

        *Keyword arguments:*

            - list_of_items -- List of items to add.
            - ordered -- Whether the list is odered (default = False).
        """
        codes = ""

        if ordered: codes += "<ol>"
        else: codes += "<ul>"
        
        for item in list_of_items:
            codes += "<li style=\"margin-left: 50\">"+item+"</li>"
        
        if ordered: codes += "</ol>"
        else: codes += "</ul>"

        self.document.append(codes)
        
    def write(self, file_name):
        """Write HTML document to file name.

        *Keyword arguments:*

            - file_name -- Complete file name to write this HTML document.
        """

        # Add footer - finalize document
        self.create_footer()

        # Writing document to file
        f = open(file_name,"w")
        for e in self.document:
            if e: f.write(e+"\n")
        f.close()

class AuxiliaryFunctions:
    """Class of auxiliary static functions."""

    @staticmethod
    def string_is_int(s):
        """Verifies if a string is a numeric integer.

        *Keyword arguments:*

            - s -- String to verify.
        """
        try:
            int(s)
            return True
        except ValueError:
            return False

    @staticmethod
    def string_is_float(s):
        """Verifies if a string is a numeric float.

        *Keyword arguments:*

            - s -- String to verify.
        """
        try:
            float(s)
            return True
        except ValueError:
            return False

    @staticmethod
    def correct_standard_bed_score(score):
        """Standardize scores between 0 and 1000.

        *Keyword arguments:*

            - score -- Score.
        """
        return min(max(score,0),1000)

    @staticmethod
    def overlap(t1, t2, strand_specific=False):
        """Checks if one interval contains any overlap with another interval.

        *Keyword arguments:*

            - t1 -- First tuple.
            - t2 -- Second tuple.
  
        *Return:*
            - -1 -- if i1 is before i2.
            - 1 -- if i1 is after i2.
            - 0 -- if there is any overlap.
        """
        if strand_specific:
            if (t1[1] <= t2[0]): return -1  # interval1 is before interval2
            if (t2[1] <= t1[0]): return 1  # interval1 is after interval2
            if t1[4] == t2[2]:
                return 0  # interval1 overlaps interval2
            else:
                return 2  # interval1 overlaps interval2 on the opposite strand
        else:
            if(t1[1] <= t2[0]): return -1 # interval1 is before interval2
            if(t2[1] <= t1[0]): return 1 # interval1 is after interval2
            return 0 # interval1 overlaps interval2

    @staticmethod
    def revcomp(s):
        """Revert complement string.

        *Keyword arguments:*

            - s -- String.
        """
        revDict = dict([("A","T"),("T","A"),("C","G"),("G","C"),("N","N")])
        return "".join([revDict[e] for e in s[::-1]])

        
def which(program):
    """Return path of program or None, see
    http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python"""
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


"""
THOR detects differential peaks in multiple ChIP-seq profiles associated
with two distinct biological conditions.

Copyright (C) 2014-2016 Manuel Allhoff (allhoff@aices.rwth-aachen.de)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

@author: Manuel Allhoff
"""

from __future__ import print_function
import sys

# from rgt.CoverageSet import CoverageSet
from rgt.GenomicRegion import GenomicRegion
from rgt.GenomicRegionSet import GenomicRegionSet
# from os.path import splitext, basename

class RegionGiver:
    regionset = GenomicRegionSet('')
    chrom_sizes_dict = {}
    
    def __init__(self, chrom_sizes, regions=None):
        self.counter = 0
        if regions is not None:
            print("Call DPs on specified regions.", file=sys.stderr)
            with open(regions) as f:
                for line in f:
                    if line:
                        line = line.strip()
                        line = line.split()
                        c, s, e = line[0], int(line[1]), int(line[2])
                #if c in contained_chrom:
                        self.regionset.add(GenomicRegion(chrom=c, initial=s, final=e))
                        self.chrom_sizes_dict[c] = e
        else:
            print("Call DPs on whole genome.", file=sys.stderr)
            with open(chrom_sizes) as f:
                for line in f:
                    line = line.strip()
                    line = line.split('\t')
                    chrom, end = line[0], int(line[1])
            #if chrom in contained_chrom:
                    self.regionset.add(GenomicRegion(chrom=chrom, initial=0, final=end))
                    self.chrom_sizes_dict[chrom] = end
        
        if not self.regionset.sequences:
            print('something wrong here', file=sys.stderr)
            sys.exit(2)
    
    def __len__(self):
        return len(self.regionset)
    
    def __iter__(self):
        for el in self.regionset:
            tmp = GenomicRegionSet('')
            tmp.add(el)
            yield tmp
        #return iter(self.regionset)
    
    def get_regionset(self):
        return self.regionset
    
    def get_chrom_dict(self):
        return self.chrom_sizes_dict
    
    
    def get_training_regionset(self):
        r = GenomicRegionSet('')
        r.add(self.regionset[self.counter])
        
        if self.counter == len(self.chrom_sizes_dict):
            return None
        else:
            self.counter += 1
            return r
        
            
            
#if regions option is set, take the values, otherwise the whole set of 
    #chromosomes as region to search for DPs
#     if test:
#         contained_chrom = ['chr1', 'chr2']
#     else:
#         #contained_chrom = get_all_chrom(bamfiles)
#         contained_chrom = ['chr1', 'chr2']



###################################################################################################
# Libraries
###################################################################################################

# Python
from os.path import basename

# Internal
from rgt.Util import ErrorHandler

# External
from Bio import motifs

###################################################################################################
# Classes
###################################################################################################


class Motif:
    """
    Represent a DNA binding affinity motif.

    Authors: Eduardo G. Gusmao.
    """

    def __init__(self, input_file_name, pseudocounts, precision, fpr, thresholds):
        """ 
        Initializes Motif.

        Variables:
        pfm -- Position Frequency Matrix.
        pwm -- Position Weight Matrix.
        pssm -- Position Specific Scoring Matrix.
        threshold -- Motif matching threshold.
        len -- Length of the motif.
        max -- Maximum PSSM score possible.
        is_palindrome -- True if consensus is biologically palindromic.
        """

        # Initializing error handler
        err = ErrorHandler()
 
        # Initializing name
        self.name = ".".join(basename(input_file_name).split(".")[:-1])
        repository = input_file_name.split("/")[-2]

        # Creating PFM & PWM
        input_file = open(input_file_name, "r")
        self.pfm = motifs.read(input_file, "pfm")
        self.pwm = self.pfm.counts.normalize(pseudocounts)
        input_file.close()
        self.len = len(self.pfm)

        # Creating PSSM
        background = {'A': 0.25, 'C': 0.25, 'G': 0.25, 'T': 0.25}
        self.pssm = self.pwm.log_odds(background)
        self.pssm_list = [self.pssm[e] for e in ["A", "C", "G", "T"]]
        self.max = self.pssm.max

        # Evaluating threshold
        try:
            if pseudocounts != 0.1 or precision != 10000:
                raise ValueError()
            self.threshold = thresholds.dict[repository][self.name][fpr]
        except Exception:
            err.throw_warning("DEFAULT_WARNING", add_msg="Parameters not matching pre-computed Fpr data. "
                                                         "Recalculating (might take a while)..")
            try:
                distribution = self.pssm.distribution(background=background, precision=precision)
            except Exception:
                err.throw_error("MM_PSEUDOCOUNT_0")
            self.threshold = distribution.threshold_fpr(fpr)

        # Evaluating if motif is palindromic
        if str(self.pfm.consensus) == str(self.pfm.consensus.reverse_complement()):
            self.is_palindrome = True
        else:
            self.is_palindrome = False


class Thresholds:
    """
    Container for all motif thresholds given default FPRs.

    Authors: Eduardo G. Gusmao.
    """

    def __init__(self, motif_data):
        """ 
        Initializes Thresholds. Motif thresholds are stored in a dictionary:
        [repository] -> [motif name] -> [fpr] -> threshold float

        Parameters:
        motif_data -- MotifData object.

        Variables:
        
        """

        # Initializing dictionary level 0
        self.dict = dict()

        # Iterating over fpr files
        for fpr_file_name in motif_data.get_fpr_list():

            # Initializing dictionary level 1
            fpr_name = ".".join(fpr_file_name.split("/")[-1].split(".")[:-1])
            self.dict[fpr_name] = dict()

            # Iterating in fpr file
            fpr_file = open(fpr_file_name, "r")
            header = fpr_file.readline()
            fpr_values = [float(e) for e in header.strip().split("\t")[1:]]
            for line in fpr_file:
                ll = line.strip().split("\t")
                # Initializing dictionary level 2
                self.dict[fpr_name][ll[0]] = dict()
                for i in range(1, len(ll)):
                    # Updating dictionary
                    self.dict[fpr_name][ll[0]][fpr_values[i-1]] = float(ll[i])
            fpr_file.close()



###################################################################################################
# Libraries
###################################################################################################

# Python
from os import remove, system, getcwd
from sys import exit
from copy import deepcopy
from optparse import SUPPRESS_HELP
import warnings

warnings.filterwarnings("ignore")

# Internal
from rgt import __version__
from ..Util import PassThroughOptionParser, ErrorHandler, HmmData, GenomeData, OverlapType
from ..ExperimentalMatrix import ExperimentalMatrix
from ..GenomicRegion import GenomicRegion
from ..GenomicRegionSet import GenomicRegionSet
from signalProcessing import GenomicSignal
from hmm import HMM
from biasTable import BiasTable
from evaluation import Evaluation
from train import TrainHMM
from evidence import Evidence
from plot import Plot

# External
import os
from numpy import array, sum, isnan
from hmmlearn.hmm import GaussianHMM
from hmmlearn import __version__ as hmm_ver

"""
HINT - HMM-based Identification of TF Footprints.
Finds transcription factor footprints given open chromatin data.

Basic Input:
- Regions (bed) in which to find footprints (i.e. enriched regions or hypersensitivity regions).
- Reads (bam) containing the open chromatin signal for DNase/ATAC and 0 <= N <= 3 histone modifications.

Dependencies:
- python >= 2.7
- numpy >= 1.4.0
- scipy >= 0.7.0
- scikit >= 0.14
- hmmlearn >= 0.1.1
- pysam >= 0.7.5
- matplotlib >= 1.6
Authors: Eduardo G. Gusmao, Manuel Allhoff, Joseph Kuo and Ivan G. Costa.
"""


def main():
    """
    Main function that performs footprint analysis.

    Keyword arguments: None

    Return: None
    """

    ###################################################################################################
    # Processing Input Arguments
    ###################################################################################################

    # Initializing ErrorHandler
    error_handler = ErrorHandler()

    # Parameters
    usage_message = ("\n--------------------------------------------------\n"
                     "The 'hint' program predicts TFBSs given open chromatin data.\n"
                     "In order to use this tools, please type: \n\n"
                     "%prog [options] <experiment_matrix>\n\n"
                     "The minimal <experiment matrix> should contain:\n"
                     "- One region file representing the regions in which the HMM\n"
                     "  will be applied. It should contain 'regions' in the type field\n"
                     "  and 'HS' in the data field\n"
                     "- One DNase-seq or ATAC-seq aligned reads file (bam) file with\n"
                     "  'reads' in the type field and 'DNASE' or 'ATAC' in the data field.\n"
                     "- Zero to Three histone modification aligned reads file (bam)\n"
                     "  with 'reads' in the type field and 'HISTONE' in the data field.\n\n"

                     "For more information, please refer to:\n"
                     "http://www.regulatory-genomics.org/hint/introduction/\n\n"

                     "For further questions or comments please refer to our group:\n"
                     "https://groups.google.com/forum/#!forum/rgtusers\n"
                     "--------------------------------------------------")
    version_message = "HINT - Regulatory Analysis Toolbox (RGT). Version: " + str(__version__)

    # Initializing Option Parser
    parser = PassThroughOptionParser(usage=usage_message, version=version_message)

    # Optional Input Options
    parser.add_option("--hmm-file", dest="hmm_file", type="string",
                      metavar="FILE_1_1[[,...,FILE_N_1];...;FILE_1_M[,...,FILE_N_M]]", default=None,
                      help=("List of HMM files separated by comma. Rules: "
                            "- If DNase/ATAC-only analysis, provide only one HMM file per group. "
                            "- If also using histone modifications: "
                            "  * If only one file is provided it will be applied for all histone data, "
                            "  * Otherwise, the list must have the same number of input histone data. "
                            "    The order of the list should be the order of the histones in the "
                            "    input_matrix file. "
                            "- In case multiple input groups are used, then multiple file lists can "
                            "  be passed using semicolon. The number of groups of lists should "
                            "  equals the number of input groups. "
                            "- If the argument is not given, then a default HMM will be used."))
    parser.add_option("--bias-table", dest="bias_table", type="string",
                      metavar="FILE1_F,FILE1_R[;...;FILEM_F,FILEM_R]", default=None,
                      help=("List of files (for each input group; separated by semicolon) with all "
                            "possible k-mers (for any k) and their bias estimates. Each input group"
                            "should have two files: one for the forward and one for the reverse strand. "
                            "Each line should contain a kmer and the bias estimate separated by tab. "
                            "Leave an empty set for histone-only analysis groups. Eg. FILE1;;FILE3."))

    # Parameters Options
    parser.add_option("--organism", dest="organism", type="string", metavar="STRING", default="hg19",
                      help=("Organism considered on the analysis. Check our full documentation for all available "
                            "options. All default files such as genomes will be based on the chosen organism "
                            "and the data.config file."))
    parser.add_option("--estimate-bias-correction", dest="estimate_bias_correction",
                      action="store_true", default=False,
                      help=("Applies DNase-seq sequence cleavage bias correction with k-mer bias estimated "
                            "from the given DNase-seq data (SLOW HINT-BC)."))
    parser.add_option("--original-regions", dest="original_regions", type="string",
                      metavar="STRING", default=None,
                      help=("The regions that used to estimate the bias table "
                            "should be bed file containing HS regions or FASTA file containing naked DNA"))
    parser.add_option("--estimate-bias-type", dest="estimate_bias_type", type="string",
                      metavar="STRING", default=None,
                      help=("The methods that used to estimate the bias table "
                            "Available options are: 'FRE' (the bias estimation is computed as the ratio "
                            "between the observed and background cleavage frequency) "
                            "and 'PWM' (frequency is replaced by pwm score)."))
    parser.add_option("--default-bias-correction", dest="default_bias_correction",
                      action="store_true", default=False,
                      help=("Applies DNase-seq cleavage bias correction with default k-mer bias "
                            "estimates (FAST HINT-BC). Please set the correct --default-bias-type "
                            "option that matches your experimental settings."))
    parser.add_option("--default-bias-type", dest="default_bias_type", type="string",
                      metavar="STRING", default="DH",
                      help=("Type of protocol used to generate the DNase-seq or ATAC-seq data. "
                            "Available options are: 'SH' (DNase-seq single-hit protocol), 'DH' "
                            "(DNase-seq double-hit protocol) and 'ATAC' (ATAC-seq data)."))

    # Output Options
    parser.add_option("--output-location", dest="output_location", type="string", metavar="PATH",
                      default=getcwd(),
                      help=("Path where the output files will be written."))
    parser.add_option("--output-fname", dest="output_fname", type="string", metavar="STRING",
                      default=None)
    parser.add_option("--print-raw-signal", dest="print_raw_signal", type="string", metavar="STRING",
                      default=None,
                      help=("If used, it will print the base overlap (raw) signals from DNase-seq "
                            " or ATAC-seq data. The option should equal the file name."
                            "The extension must be (.wig)."))
    parser.add_option("--print-bc-signal", dest="print_bc_signal", type="string", metavar="STRING",
                      default=None,
                      help=("If used, it will print the DNase-seq or ATAC-seq bias-corrected signal. "
                            "The option should equal the file name. The extension must be (.wig)."))
    parser.add_option("--print-norm-signal", dest="print_norm_signal", type="string", metavar="STRING",
                      default=None,
                      help=("If used, it will print the normalized signals from DNase-seq "
                            " or ATAC-seq data. The option should equal the file name."
                            "The extension must be (.wig)."))
    parser.add_option("--print-slope-signal", dest="print_slope_signal", type="string", metavar="STRING",
                      default=None,
                      help=("If used, it will print the slope signals from DNase-seq "
                            " or ATAC-seq data. The option should equal the file name."
                            "The extension must be (.wig)."))
    parser.add_option("--print-line-plot", dest="print_line_plot",
                      action="store_true", default=False,
                      help=("If used, it will print the line plot of raw signal and bias corrected"
                            "signal for the particular motif."))
    parser.add_option("--window-size", dest="window_size", type="int", metavar="INT", default=50)
    parser.add_option("--motif-file", dest="motif_file", type="string", metavar="STRING",
                      default=None,
                      help=("A bed file containing all motif-predicted binding sites (MPBSs)."
                            "The extension must be (.bed)."))
    parser.add_option("--motif-name", dest="motif_name", type="string", metavar="STRING",
                      default=None)
    parser.add_option("--protection-score", dest="protection_score",
                      action="store_true", default=False,
                      help=("If used, it will print the protection score"))
    parser.add_option("--strands-specific", dest="strands_specific",
                      action="store_true", default=False,
                      help=("A boolean indicating if DNaseI digestion site "
                            "counts should be aggregated across strands. "
                            "default: False"))

    # Train Options
    parser.add_option("--train-hmm", dest="train_hmm",
                      action="store_true", default=False,
                      help=("If used, HINT will train a hidden Markov model (HMM) based on "
                            "the annotation data"))
    parser.add_option("--bam-file", dest="bam_file", type="string", metavar="STRING",
                      default=None,
                      help=("A bam file containing all the DNase-seq reads."))
    parser.add_option("--annotate-file", dest="annotate_file", type="string", metavar="STRING",
                      default=None,
                      help=("A annotate file containing all the states."))
    parser.add_option("--print-bed-file", dest="print_bed_file",
                      action="store_true", default=False,
                      help=("If used, HINT will output the bed file containing "
                            "the annotated regions, so that you can visualize "
                            "you HMM annotation for potential errors"))
    parser.add_option("--model-fname", dest="model_fname", type="string", metavar="STRING",
                      default="model",
                      help=("The output file name"))

    # Evaluation Options
    parser.add_option("--evaluate-footprints", dest="evaluate_footprints",
                      action="store_true", default=False,
                      help=("If used, HINT will evaluate the footprints prediction."))
    parser.add_option("--tf-name", dest="tf_name", type="string", metavar="STRING",
                      default=None,
                      help=("The name of transcription factor"))
    parser.add_option("--tfbs-file", dest="tfbs_file", type="string", metavar="STRING",
                      default=None,
                      help=("A bed file containing all motif-predicted binding sites (MPBSs)."
                            "The values in the bed SCORE field will be used to rank the MPBSs."
                            "The extension must be (.bed)."))
    parser.add_option("--footprint-file", dest="footprint_file", type="string",
                      metavar="FILE1,FILE2,FILE3,FILE4...",
                      default=None,
                      help=("The bed files containing the footprint predictions."
                            "The extension must be (.bed)."))
    parser.add_option("--footprint-name", dest="footprint_name", type="string",
                      metavar="NAME1,NAME2,NAME3,NAME4...",
                      default=None,
                      help=("The methods name used to predicted the footprint."
                            "The number of methods name must be consistent with that of footprint file"))
    parser.add_option("--alignment-file", dest="alignment_file", type="string",
                      metavar="FILE",
                      default=None,
                      help=("A bam file containing all the DNase-seq reads."
                            "Used to fetch the tag counts for SEG (segmentation approach)"))
    parser.add_option("--footprint-type", dest="footprint_type", type="string",
                      metavar="TYPE1,TYPE2,TYPE3,TYPE4...",
                      default=None,
                      help=("The methods type used to predicted the footprint."
                            "Must be SC (site centric) or SEG (segmentation approach)"))
    parser.add_option("--print-roc-curve", dest="print_roc_curve",
                      action="store_true", default=False,
                      help=("If used, HINT will print the receiver operating characteristic curve."))
    parser.add_option("--print-pr-curve", dest="print_pr_curve",
                      action="store_true", default=False,
                      help=("If used, HINT will print the precision recall curve."))

    # Evidence Options
    parser.add_option("--create-evidence", dest="create_evidence",
                      action="store_true", default=False,
                      help=("If used, HINT will create a bed file with MPBSs with and "
                            "without evidence. Also, the name of the instances will be Y "
                            "for evidence, N for non evidence."))
    parser.add_option("--peak-ext", dest="peak_ext", type="int", metavar="INT", default=100,
                      help=("The number used to extend the ChIP-seq summit"))
    parser.add_option("--mpbs-name", dest="mpbs_name", type="string",
                      default=None,
                      help=("motif predicted binding sites name"))
    parser.add_option("--tfbs-summit-fname", dest="tfbs_summit_fname", type="string",
                      default=None,
                      help=("the ChIP-seq peak files"))
    parser.add_option("--mpbs-fname", dest="mpbs_fname", type="string",
                      default=None,
                      help=("motif predicted binding sites file"))

    # GENERAL Hidden Options
    parser.add_option("--region-total-ext", dest="region_total_ext", type="int", metavar="INT", default=10000,
                      help=SUPPRESS_HELP)
    parser.add_option("--fp-limit-size", dest="fp_limit_size", type="int", metavar="INT", default=50,
                      help=SUPPRESS_HELP)
    parser.add_option("--fp-limit-size-histone", dest="fp_limit_size_histone", type="int",
                      metavar="INT", default=2000, help=SUPPRESS_HELP)
    parser.add_option("--fp-limit-size-ext", dest="fp_limit_size_ext", type="int", metavar="INT", default=10,
                      help=SUPPRESS_HELP)
    parser.add_option("--fp-limit-size-ext-histone", dest="fp_limit_size_ext_histone",
                      type="int", metavar="INT", default=200, help=SUPPRESS_HELP)
    parser.add_option("--fp-ext", dest="fp_ext", type="int", metavar="INT", default=5,
                      help=SUPPRESS_HELP)
    parser.add_option("--fp-ext-histone", dest="fp_ext_histone", type="int", metavar="INT", default=50,
                      help=SUPPRESS_HELP)
    parser.add_option("--tc-ext", dest="tc_ext", type="int", metavar="INT", default=50,
                      help=SUPPRESS_HELP)
    parser.add_option("--tc-ext-histone", dest="tc_ext_histone", type="int", metavar="INT", default=500,
                      help=SUPPRESS_HELP)

    # DNASE Hidden Options
    parser.add_option("--dnase-initial-clip", dest="dnase_initial_clip", type="int",
                      metavar="INT", default=1000, help=SUPPRESS_HELP)
    parser.add_option("--dnase-sg-window-size", dest="dnase_sg_window_size", type="int",
                      metavar="INT", default=9, help=SUPPRESS_HELP)
    parser.add_option("--dnase-norm-per", dest="dnase_norm_per", type="float",
                      metavar="INT", default=98, help=SUPPRESS_HELP)
    parser.add_option("--dnase-slope-per", dest="dnase_slope_per", type="float",
                      metavar="INT", default=98, help=SUPPRESS_HELP)
    parser.add_option("--dnase-downstream-ext", dest="dnase_downstream_ext", type="int",
                      metavar="INT", default=1, help=SUPPRESS_HELP)
    parser.add_option("--dnase-upstream-ext", dest="dnase_upstream_ext", type="int",
                      metavar="INT", default=0, help=SUPPRESS_HELP)
    parser.add_option("--dnase-forward-shift", dest="dnase_forward_shift", type="int",
                      metavar="INT", default=0, help=SUPPRESS_HELP)
    parser.add_option("--dnase-reverse-shift", dest="dnase_reverse_shift", type="int",
                      metavar="INT", default=0, help=SUPPRESS_HELP)
    parser.add_option("--dnase-bias-correction-k", dest="dnase_bias_correction_k", type="int",
                      metavar="INT", default=6, help=SUPPRESS_HELP)

    # ATAC Hidden Options
    parser.add_option("--atac-initial-clip", dest="atac_initial_clip", type="int",
                      metavar="INT", default=1000, help=SUPPRESS_HELP)
    parser.add_option("--atac-sg-window-size", dest="atac_sg_window_size", type="int",
                      metavar="INT", default=9, help=SUPPRESS_HELP)
    parser.add_option("--atac-norm-per", dest="atac_norm_per", type="float",
                      metavar="INT", default=98, help=SUPPRESS_HELP)
    parser.add_option("--atac-slope-per", dest="atac_slope_per", type="float",
                      metavar="INT", default=98, help=SUPPRESS_HELP)
    parser.add_option("--atac-downstream-ext", dest="atac_downstream_ext", type="int",
                      metavar="INT", default=1, help=SUPPRESS_HELP)
    parser.add_option("--atac-upstream-ext", dest="atac_upstream_ext", type="int",
                      metavar="INT", default=0, help=SUPPRESS_HELP)
    parser.add_option("--atac-forward-shift", dest="atac_forward_shift", type="int",
                      metavar="INT", default=5, help=SUPPRESS_HELP)
    parser.add_option("--atac-reverse-shift", dest="atac_reverse_shift", type="int",
                      metavar="INT", default=-4, help=SUPPRESS_HELP)
    parser.add_option("--atac-bias-correction-k", dest="atac_bias_correction_k", type="int",
                      metavar="INT", default=6, help=SUPPRESS_HELP)

    # HISTONE Hidden Options
    parser.add_option("--histone-initial-clip", dest="histone_initial_clip", type="int",
                      metavar="INT", default=1000, help=SUPPRESS_HELP)
    parser.add_option("--histone-sg-window-size", dest="histone_sg_window_size", type="int",
                      metavar="INT", default=201, help=SUPPRESS_HELP)
    parser.add_option("--histone-norm-per", dest="histone_norm_per", type="float",
                      metavar="INT", default=98, help=SUPPRESS_HELP)
    parser.add_option("--histone-slope-per", dest="histone_slope_per", type="float",
                      metavar="INT", default=98, help=SUPPRESS_HELP)
    parser.add_option("--histone-downstream-ext", dest="histone_downstream_ext", type="int",
                      metavar="INT", default=200, help=SUPPRESS_HELP)
    parser.add_option("--histone-upstream-ext", dest="histone_upstream_ext", type="int",
                      metavar="INT", default=0, help=SUPPRESS_HELP)
    parser.add_option("--histone-forward-shift", dest="histone_forward_shift", type="int",
                      metavar="INT", default=0, help=SUPPRESS_HELP)
    parser.add_option("--histone-reverse-shift", dest="histone_reverse_shift", type="int",
                      metavar="INT", default=0, help=SUPPRESS_HELP)

    # Processing Options
    options, arguments = parser.parse_args()
    if len(arguments) < 1:
        print(usage_message)
        exit(1)
    # if(not arguments or len(arguments) > 1): error_handler.throw_error("FP_WRONG_ARGUMENT")

    # General hidden options ###############################################################
    region_total_ext = options.region_total_ext
    fp_limit_size = options.fp_limit_size
    fp_limit_size_histone = options.fp_limit_size_histone
    fp_limit_size_ext = options.fp_limit_size_ext
    fp_limit_size_ext_histone = options.fp_limit_size_ext_histone
    fp_ext = options.fp_ext
    fp_ext_histone = options.fp_ext_histone
    tc_ext = options.tc_ext
    tc_ext_histone = options.tc_ext_histone
    # DNASE Hidden Options
    dnase_initial_clip = options.dnase_initial_clip
    dnase_sg_window_size = options.dnase_sg_window_size
    dnase_norm_per = options.dnase_norm_per
    dnase_slope_per = options.dnase_slope_per
    dnase_downstream_ext = options.dnase_downstream_ext
    dnase_upstream_ext = options.dnase_upstream_ext
    dnase_forward_shift = options.dnase_forward_shift
    dnase_reverse_shift = options.dnase_reverse_shift
    dnase_bias_correction_k = options.dnase_bias_correction_k
    # ATAC Hidden Options
    atac_initial_clip = options.atac_initial_clip
    atac_sg_window_size = options.atac_sg_window_size
    atac_norm_per = options.atac_norm_per
    atac_slope_per = options.atac_slope_per
    atac_downstream_ext = options.atac_downstream_ext
    atac_upstream_ext = options.atac_upstream_ext
    atac_forward_shift = options.atac_forward_shift
    atac_reverse_shift = options.atac_reverse_shift
    atac_bias_correction_k = options.atac_bias_correction_k
    # HISTONE Hidden Options
    histone_initial_clip = options.histone_initial_clip
    histone_sg_window_size = options.histone_initial_clip
    histone_norm_per = options.histone_norm_per
    histone_slope_per = options.histone_slope_per
    histone_downstream_ext = options.histone_downstream_ext
    histone_upstream_ext = options.histone_upstream_ext
    histone_forward_shift = options.histone_forward_shift
    histone_reverse_shift = options.histone_reverse_shift

    ########################################################################################

    ##########################################################################################

    # IF HINT is required to output the line plot and motif logo
    if options.print_line_plot:
        plot = Plot(options.bam_file, options.motif_file, options.motif_name, options.window_size,
                    atac_downstream_ext, atac_upstream_ext, atac_forward_shift, atac_reverse_shift,
                    atac_initial_clip, options.organism, options.bias_table,
                    atac_bias_correction_k, options.protection_score, options.strands_specific,
                    options.output_location)
        plot.line()
        return

    # If HINT is required to create the validation data set
    if options.create_evidence:
        evidence = Evidence(options.peak_ext, options.mpbs_name, options.tfbs_summit_fname,
                            options.mpbs_fname, options.output_location)
        evidence.create_file()
        return

    # If HINT is required to evaluate the existing footprint predictions
    if options.evaluate_footprints:
        evaluation = Evaluation(options.tf_name, options.tfbs_file, options.footprint_file,
                                options.footprint_name, options.footprint_type,
                                options.print_roc_curve, options.print_roc_curve,
                                options.output_location, options.alignment_file, options.organism)
        evaluation.chip_evaluate()
        return

    # If HINT is required to train a hidden Markov model (HMM)
    if options.train_hmm:
        train_hmm_model = TrainHMM(options.bam_file, options.annotate_file, options.print_bed_file,
                                   options.output_location, options.model_fname,
                                   options.print_raw_signal, options.print_bc_signal,
                                   options.print_norm_signal, options.print_slope_signal,
                                   atac_initial_clip, atac_downstream_ext, atac_upstream_ext,
                                   atac_forward_shift, atac_reverse_shift,
                                   options.estimate_bias_correction, options.estimate_bias_type,
                                   options.bias_table,
                                   options.original_regions, options.organism,
                                   atac_bias_correction_k)
        train_hmm_model.train()
        return

    ########################################################################################################
    # Output wig signal
    if (options.print_raw_signal):
        system("touch " + options.print_raw_signal + " | echo -n "" > " + options.print_raw_signal)
    if (options.print_bc_signal):
        system("touch " + options.print_bc_signal + " | echo -n "" > " + options.print_bc_signal)
    if (options.print_norm_signal):
        system("touch " + options.print_norm_signal + " | echo -n "" > " + options.print_norm_signal)
    if (options.print_slope_signal):
        system("touch " + options.print_raw_signal + " | echo -n "" > " + options.print_slope_signal)

    # Global class initialization
    genome_data = GenomeData(options.organism)
    hmm_data = HmmData()

    ###################################################################################################
    # Reading Input Matrix
    ###################################################################################################

    # Reading input argument
    input_matrix = arguments[0]

    # Create experimental matrix
    try:
        exp_matrix = ExperimentalMatrix()
        exp_matrix.read(input_matrix)
    except Exception:
        error_handler.throw_error("FP_WRONG_EXPMAT")

    ###################################################################################################
    # Reading Input
    ###################################################################################################

    # Group class
    class Group:
        def __init__(self):
            self.name = None
            self.original_regions = None
            self.regions = None
            self.dnase_file = None
            self.histone_file_list = []
            self.dnase_only = True
            self.histone_only = True
            self.hmm = []
            self.flag_multiple_hmms = False
            self.bias_table = None
            self.is_atac = False

    # Initialization
    name_list = exp_matrix.names
    type_list = exp_matrix.types
    file_dict = exp_matrix.files
    fields_dict = exp_matrix.fieldsDict
    objects_dict = exp_matrix.objectsDict

    # Populating fields dict data
    for e in ["HS", "DNASE", "ATAC", "HISTONE"]:
        try:
            fields_dict["data"][e]
        except Exception:
            fields_dict["data"][e] = []

    # Fetching chromosome sizes
    chrom_sizes_file_name = genome_data.get_chromosome_sizes()
    chrom_sizes_file = open(chrom_sizes_file_name, "r")
    chrom_sizes_dict = dict()
    for chrom_sizes_entry_line in chrom_sizes_file:
        chrom_sizes_entry_vec = chrom_sizes_entry_line.strip().split("\t")
        chrom_sizes_dict[chrom_sizes_entry_vec[0]] = int(chrom_sizes_entry_vec[1])
    chrom_sizes_file.close()

    # Fetching files per group
    group_list = []
    for g in fields_dict["group"].keys():
        group = Group()
        group.name = g
        for i in range(0, len(fields_dict["group"][g])):
            if (name_list[i] in fields_dict["data"]["HS"]):
                group.original_regions = objects_dict[name_list[i]]
                group.regions = deepcopy(group.original_regions)
                group.regions.extend(int(region_total_ext / 2), int(region_total_ext / 2))  # Extending
                group.regions.merge()  # Sort & Merge
            elif (name_list[i] in fields_dict["data"]["DNASE"]):
                group.dnase_file = GenomicSignal(file_dict[name_list[i]])
                group.dnase_file.load_sg_coefs(dnase_sg_window_size)
                group.is_atac = False
            elif (name_list[i] in fields_dict["data"]["ATAC"]):
                group.dnase_file = GenomicSignal(file_dict[name_list[i]])
                group.dnase_file.load_sg_coefs(atac_sg_window_size)
                group.is_atac = True
            elif (name_list[i] in fields_dict["data"]["HISTONE"]):
                group.histone_file_list.append(GenomicSignal(file_dict[name_list[i]]))
                group.histone_file_list[-1].load_sg_coefs(histone_sg_window_size)
            else:
                pass  # TODO ERROR (Category of data outside "HS, DNASE, ATAC, HISTONE")
        if (group.dnase_file): group.histone_only = False
        if (group.histone_file_list): group.dnase_only = False
        if (group.histone_only and group.dnase_only): pass  # TODO ERROR (There is no DNase or histone data)
        if (not group.original_regions): pass  # TODO ERROR (There is no HS regions)
        group_list.append(group)

    ###################################################################################################
    # Fetching Bias Table
    ###################################################################################################

    bias_correction = False
    if (options.bias_table):

        bias_table_group_list = options.bias_table.split(";")
        if (len(bias_table_group_list) != len(group_list)): pass  # TODO ERROR
        for g in range(0, len(group_list)):
            group = group_list[g]
            bias_table_list = bias_table_group_list[g].split(",")
            if (group.histone_only): continue
            group.bias_table = BiasTable().load_table(table_file_name_F=bias_table_list[0],
                                                      table_file_name_R=bias_table_list[1])
        bias_correction = True

    elif (options.estimate_bias_correction):
        for group in group_list:
            if (group.histone_only): continue
            if (group.is_atac):
                my_k_nb = atac_bias_correction_k
                my_shift = atac_downstream_ext
                group.bias_table = BiasTable().estimate_table(regions=group.original_regions,
                                                              dnase_file_name=group.dnase_file.file_name,
                                                              genome_file_name=genome_data.get_genome(),
                                                              k_nb=my_k_nb,
                                                              forward_shift=atac_forward_shift,
                                                              reverse_shift=atac_reverse_shift)
            else:
                my_k_nb = dnase_bias_correction_k
                my_shift = dnase_downstream_ext
                group.bias_table = BiasTable().estimate_table(regions=group.original_regions,
                                                              dnase_file_name=group.dnase_file.file_name,
                                                              genome_file_name=genome_data.get_genome(),
                                                              k_nb=my_k_nb,
                                                              forward_shift=dnase_forward_shift,
                                                              reverse_shift=dnase_reverse_shift)
        bias_correction = True

    elif (options.default_bias_correction):

        for group in group_list:
            if (group.histone_only): continue

            if (options.default_bias_type == "SH"):
                my_table_file_F = hmm_data.get_default_bias_table_F_SH()
                my_table_file_R = hmm_data.get_default_bias_table_R_SH()
            elif (options.default_bias_type == "DH"):
                my_table_file_F = hmm_data.get_default_bias_table_F_DH()
                my_table_file_R = hmm_data.get_default_bias_table_R_DH()
            else:
                my_table_file_F = hmm_data.get_default_bias_table_F_ATAC()
                my_table_file_R = hmm_data.get_default_bias_table_R_ATAC()

            group.bias_table = BiasTable().load_table(table_file_name_F=my_table_file_F,
                                                      table_file_name_R=my_table_file_R)
        bias_correction = True
    ###################################################################################################
    # Creating HMMs
    ###################################################################################################

    # Fetching HMM input
    flag_multiple_hmms = False
    if (options.hmm_file):  # Argument is passed

        hmm_group_list = options.hmm_file.split(";")
        if (len(hmm_group_list) != len(group_list)): pass  # TODO ERROR
        for g in range(0, len(group_list)):

            group = group_list[g]

            # Fetching list of HMM files
            group.hmm = hmm_group_list[g].split(",")

            # Verifying HMM application mode (one HMM or multiple HMM files)
            if (len(group.hmm) == 1):
                group.flag_multiple_hmms = False
                group.hmm = group.hmm[0]
            elif (len(group.hmm) == len(histone_file_name_list)):
                flag_multiple_hmms = True
            else:
                error_handler.throw_error("FP_NB_HMMS")

    else:  # Argument was not passed

        for group in group_list:

            group.flag_multiple_hmms = False
            if (group.dnase_only):
                if (bias_correction):
                    if (group.is_atac):
                        group.hmm = hmm_data.get_default_hmm_atac_bc()
                    else:
                        group.hmm = hmm_data.get_default_hmm_dnase_bc()
                else:
                    if (group.is_atac):
                        group.hmm = hmm_data.get_default_hmm_atac()
                    else:
                        group.hmm = hmm_data.get_default_hmm_dnase()
            elif (group.histone_only):
                group.hmm = hmm_data.get_default_hmm_histone()
            else:
                if (bias_correction):
                    if (group.is_atac):
                        group.hmm = hmm_data.get_default_hmm_atac_histone_bc()
                    else:
                        group.hmm = hmm_data.get_default_hmm_dnase_histone_bc()
                else:
                    if (group.is_atac):
                        group.hmm = hmm_data.get_default_hmm_atac_histone()
                    else:
                        group.hmm = hmm_data.get_default_hmm_dnase_histone()

    # Creating scikit HMM list
    for group in group_list:

        if (group.flag_multiple_hmms):

            hmm_list = []
            for hmm_file_name in group.hmm:

                try:
                    hmm_scaffold = HMM()
                    hmm_scaffold.load_hmm(hmm_file_name)
                    if (int(hmm_ver.split(".")[0]) <= 0 and int(hmm_ver.split(".")[1]) <= 1):
                        scikit_hmm = GaussianHMM(n_components=hmm_scaffold.states, covariance_type="full",
                                                 transmat=array(hmm_scaffold.A), startprob=array(hmm_scaffold.pi))
                        scikit_hmm.means_ = array(hmm_scaffold.means)
                        scikit_hmm.covars_ = array(hmm_scaffold.covs)
                    else:
                        scikit_hmm = GaussianHMM(n_components=hmm_scaffold.states, covariance_type="full")
                        scikit_hmm.startprob_ = array(hmm_scaffold.pi)
                        scikit_hmm.transmat_ = array(hmm_scaffold.A)
                        scikit_hmm.means_ = array(hmm_scaffold.means)
                        scikit_hmm.covars_ = array(hmm_scaffold.covs)

                except Exception:
                    error_handler.throw_error("FP_HMM_FILES")
                hmm_list.append(scikit_hmm)

            group.hmm = hmm_list

        else:

            scikit_hmm = None
            try:
                hmm_scaffold = HMM()
                hmm_scaffold.load_hmm(group.hmm)
                if (int(hmm_ver.split(".")[0]) <= 0 and int(hmm_ver.split(".")[1]) <= 1):
                    scikit_hmm = GaussianHMM(n_components=hmm_scaffold.states, covariance_type="full",
                                             transmat=array(hmm_scaffold.A), startprob=array(hmm_scaffold.pi))
                    scikit_hmm.means_ = array(hmm_scaffold.means)
                    scikit_hmm.covars_ = array(hmm_scaffold.covs)
                else:
                    scikit_hmm = GaussianHMM(n_components=hmm_scaffold.states, covariance_type="full")
                    scikit_hmm.startprob_ = array(hmm_scaffold.pi)
                    scikit_hmm.transmat_ = array(hmm_scaffold.A)
                    scikit_hmm.means_ = array(hmm_scaffold.means)
                    scikit_hmm.covars_ = array(hmm_scaffold.covs)


            except Exception:
                error_handler.throw_error("FP_HMM_FILES")
            group.hmm = scikit_hmm

    ###################################################################################################
    # Main Pipeline
    ###################################################################################################

    # Iterating over groups
    for group in group_list:

        # Initializing result set
        footprints = GenomicRegionSet(group.name)

        # Iterating over regions
        for r in group.regions.sequences:

            ###################################################################################################
            # DNASE ONLY
            ###################################################################################################

            if (group.dnase_only):

                # Fetching DNase signal
                try:
                    if (group.is_atac):
                        dnase_norm, dnase_slope = group.dnase_file.get_signal(r.chrom, r.initial, r.final,
                                                                              atac_downstream_ext, atac_upstream_ext,
                                                                              atac_forward_shift, atac_reverse_shift,
                                                                              atac_initial_clip, atac_norm_per,
                                                                              atac_slope_per,
                                                                              group.bias_table,
                                                                              genome_data.get_genome(),
                                                                              options.print_raw_signal,
                                                                              options.print_bc_signal,
                                                                              options.print_norm_signal,
                                                                              options.print_slope_signal)
                    else:
                        dnase_norm, dnase_slope = group.dnase_file.get_signal(r.chrom, r.initial, r.final,
                                                                              dnase_downstream_ext, dnase_upstream_ext,
                                                                              dnase_forward_shift, dnase_reverse_shift,
                                                                              dnase_initial_clip, dnase_norm_per,
                                                                              dnase_slope_per,
                                                                              group.bias_table,
                                                                              genome_data.get_genome(),
                                                                              options.print_raw_signal,
                                                                              options.print_bc_signal,
                                                                              options.print_norm_signal,
                                                                              options.print_slope_signal)
                except Exception:
                    raise
                    error_handler.throw_warning("FP_DNASE_PROC", add_msg="for region (" + ",".join([r.chrom,
                                                                                                    str(r.initial), str(
                            r.final)]) + "). This iteration will be skipped.")
                    continue

                # Formatting sequence
                try:
                    input_sequence = array([dnase_norm, dnase_slope]).T
                except Exception:
                    raise
                    error_handler.throw_warning("FP_SEQ_FORMAT", add_msg="for region (" + ",".join([r.chrom,
                                                                                                    str(r.initial), str(
                            r.final)]) + "). This iteration will be skipped.")
                    continue

                # Applying HMM
                if (isinstance(group.hmm, list)): continue  # TODO ERROR
                if (isnan(sum(input_sequence))): continue  # Handling NAN's in signal / hmmlearn throws error TODO ERROR
                try:
                    posterior_list = group.hmm.predict(input_sequence)
                except Exception:
                    raise
                    error_handler.throw_warning("FP_HMM_APPLIC", add_msg="in region (" + ",".join([r.chrom,
                                                                                                   str(r.initial), str(
                            r.final)]) + "). This iteration will be skipped.")
                    continue

                # Formatting results
                start_pos = 0
                flag_start = False
                fp_state_nb = 4
                for k in range(r.initial, r.initial + len(posterior_list)):
                    curr_index = k - r.initial
                    if (flag_start):
                        if (posterior_list[curr_index] != fp_state_nb):
                            if (k - start_pos < fp_limit_size):
                                fp = GenomicRegion(r.chrom, start_pos, k)
                                footprints.add(fp)
                            flag_start = False
                    else:
                        if (posterior_list[curr_index] == fp_state_nb):
                            flag_start = True
                            start_pos = k
                if (flag_start):
                    fp = GenomicRegion(r.chrom, start_pos, r.final)
                    footprints.add(fp)

            ###################################################################################################
            # HISTONES
            ###################################################################################################

            else:

                # Fetching DNase signal
                if (not group.histone_only):
                    try:
                        if (group.is_atac):
                            dnase_norm, dnase_slope = group.dnase_file.get_signal(r.chrom, r.initial, r.final,
                                                                                  atac_downstream_ext,
                                                                                  atac_upstream_ext,
                                                                                  atac_forward_shift,
                                                                                  atac_reverse_shift,
                                                                                  dnase_initial_clip, dnase_norm_per,
                                                                                  dnase_slope_per,
                                                                                  group.bias_table,
                                                                                  genome_data.get_genome(),
                                                                                  options.print_raw_signal,
                                                                                  options.print_bc_signal,
                                                                                  options.print_norm_signal,
                                                                                  options.print_slope_signal)
                        else:
                            dnase_norm, dnase_slope = group.dnase_file.get_signal(r.chrom, r.initial, r.final,
                                                                                  dnase_downstream_ext,
                                                                                  dnase_upstream_ext,
                                                                                  dnase_forward_shift,
                                                                                  dnase_reverse_shift,
                                                                                  dnase_initial_clip, dnase_norm_per,
                                                                                  dnase_slope_per,
                                                                                  group.bias_table,
                                                                                  genome_data.get_genome(),
                                                                                  options.print_raw_signal,
                                                                                  options.print_bc_signal,
                                                                                  options.print_norm_signal,
                                                                                  options.print_slope_signal)
                    except Exception:
                        raise
                        error_handler.throw_warning("FP_DNASE_PROC", add_msg="for region (" + ",".join([r.chrom,
                                                                                                        str(r.initial),
                                                                                                        str(
                                                                                                            r.final)]) + "). This iteration will be skipped.")
                        continue

                # Iterating over histone modifications
                for i in range(0, len(group.histone_file_list)):

                    # Fetching histone signal
                    try:
                        histone_file = group.histone_file_list[i]
                        histone_norm, histone_slope = histone_file.get_signal(r.chrom, r.initial, r.final,
                                                                              histone_downstream_ext,
                                                                              histone_upstream_ext,
                                                                              histone_forward_shift,
                                                                              histone_reverse_shift,
                                                                              histone_initial_clip, histone_norm_per,
                                                                              histone_slope_per, False, False, False,
                                                                              False, False)
                    except Exception:
                        raise
                        error_handler.throw_warning("FP_HISTONE_PROC", add_msg="for region (" + ",".join([r.chrom,
                                                                                                          str(
                                                                                                              r.initial),
                                                                                                          str(
                                                                                                              r.final)]) + ") and histone modification " + histone_file.file_name + ". This iteration will be skipped for this histone.")
                        continue

                    # Formatting sequence
                    try:
                        if (group.histone_only):
                            input_sequence = array([histone_norm, histone_slope]).T
                        else:
                            input_sequence = array([dnase_norm, dnase_slope, histone_norm, histone_slope]).T
                    except Exception:
                        raise
                        error_handler.throw_warning("FP_SEQ_FORMAT", add_msg="for region (" + ",".join(
                            [r.chrom, str(r.initial), str(
                                r.final)]) + ") and histone modification " + histone_file.file_name + ". This iteration will be skipped.")
                        continue

                    # Applying HMM
                    if (flag_multiple_hmms):
                        current_hmm = group.hmm[i]
                    else:
                        current_hmm = group.hmm
                    if (
                            isnan(sum(
                                input_sequence))): continue  # Handling NAN's in signal / hmmlearn throws error TODO ERROR
                    try:
                        posterior_list = current_hmm.predict(input_sequence)
                    except Exception:
                        raise
                        error_handler.throw_warning("FP_HMM_APPLIC", add_msg="in region (" + ",".join(
                            [r.chrom, str(r.initial), str(
                                r.final)]) + ") and histone modification " + histone_file.file_name + ". This iteration will be skipped.")
                        continue

                    # Histone-only limit size
                    if (group.histone_only):
                        fp_limit_size = fp_limit_size_histone
                        fp_state_nb = 4
                    else:
                        fp_state_nb = 7

                    # Formatting results
                    start_pos = 0
                    flag_start = False
                    for k in range(r.initial, r.initial + len(posterior_list)):
                        curr_index = k - r.initial
                        if (flag_start):
                            if (posterior_list[curr_index] != fp_state_nb):
                                if (k - start_pos < fp_limit_size):
                                    fp = GenomicRegion(r.chrom, start_pos, k)
                                    footprints.add(fp)
                                flag_start = False
                        else:
                            if (posterior_list[curr_index] == fp_state_nb):
                                flag_start = True
                                start_pos = k
                    if (flag_start):
                        fp = GenomicRegion(r.chrom, start_pos, r.final)
                        footprints.add(fp)

        ###################################################################################################
        # Post-processing
        ###################################################################################################

        # Parameters
        if (group.histone_only):
            fp_limit = fp_limit_size_ext_histone
            fp_ext = fp_ext_histone
            tc_ext = tc_ext_histone
            tcsignal = group.histone_file_list[0]
            tcext1 = histone_downstream_ext
            tcext2 = histone_upstream_ext
            tcshift1 = histone_forward_shift
            tcshift2 = histone_reverse_shift
            tcinitialclip = histone_initial_clip
        else:
            fp_limit = fp_limit_size_ext
            fp_ext = fp_ext
            tc_ext = tc_ext
            tcsignal = group.dnase_file
            if (group.is_atac):
                tcext1 = atac_downstream_ext
                tcext2 = atac_upstream_ext
                tcshift1 = atac_forward_shift
                tcshift2 = atac_reverse_shift
                tcinitialclip = atac_initial_clip
            else:
                tcext1 = dnase_downstream_ext
                tcext2 = dnase_upstream_ext
                tcshift1 = dnase_forward_shift
                tcshift2 = dnase_reverse_shift
                tcinitialclip = dnase_initial_clip

        # Sorting and Merging
        footprints.merge()

        # Overlapping results with original regions
        footprints = footprints.intersect(group.original_regions, mode=OverlapType.ORIGINAL)

        # Extending footprints
        for f in footprints.sequences:
            if (f.final - f.initial < fp_limit):
                f.initial = max(0, f.initial - fp_ext)
                f.final = f.final + fp_ext
            if (f.final - f.initial > 2 * fp_limit):
                mid = (f.initial + f.final) / 2
                f.initial = max(mid - fp_limit, 0)
                f.final = f.final + fp_limit

        # Evaluating TC
        for f in footprints.sequences:
            mid = (f.initial + f.final) / 2
            p1 = max(mid - tc_ext, 0)
            p2 = min(mid + tc_ext, chrom_sizes_dict[f.chrom])
            try:
                tag_count = tcsignal.get_tag_count(f.chrom, p1, p2, tcext1, tcext2, tcshift1, tcshift2, tcinitialclip)
            except Exception:
                tag_count = 0
            f.data = str(int(tag_count))

        ###################################################################################################
        # Writing output
        ###################################################################################################

        # Creating output file
        # output_file_name = options.output_location + options.output_fname + ".bed"
        output_file_name = os.path.join(options.output_location, "{}.bed".format(options.output_fname))
        footprints.write_bed(output_file_name)



from rgt.AnnotationSet import AnnotationSet

gene_source = "/home/egg/Projects/SVN/reg-gen/data/hg19/teste.gtf"
#gene_source = "/home/egg/Projects/SVN/reg-gen/data/hg19/gencode_annotation.gtf"

tf_source = "/home/egg/Projects/SVN/reg-gen/data/motifs/test.mtf"

ann_set = AnnotationSet(gene_source,tf_source)






"""
GenomicRegion
===================
GenomicRegion describes a genomic region.

"""
class GenomicRegion:
    """*Keyword arguments:*

            - chrom -- Chromosome.
            - initial -- Start position
            - final -- End position
            - name -- Name of the region
            - orientation -- Orientation of the region, "+" or "-"
            - data -- Extra information
            - proximity -- Close genes
    """
    #__slots__ = ['chrom', 'initial', 'final', 'name', 'orientation', 'data', 'proximity']

    def __init__(self, chrom, initial, final, name=None, orientation=None, data=None, proximity = None):
        self.chrom = str(chrom) #chrom should be a string, not an integer
        if not isinstance( initial, int ) or not isinstance(final, int):
            raise ValueError('The initial and final input for GenomicRegion should be integer.')
        self.initial = initial
        self.final = final
        self.name = name
        self.orientation = orientation
        self.data = data # data can be integer, string, list
        self.proximity = proximity

    def get_data(self, as_list=False):
        """Return data as string (with special separating character (_$_)) or as list.

        *Keyword arguments:*
        
            - as_list -- Return a list instead of a string.
        """
        if not as_list:
            return self.data
        else:
            tmp = self.data.split("_$_")
            return tmp

    def __len__(self):
        """Return length of GenomicRegion."""
        return self.final - self.initial

    def __str__(self):
        """Give informal string representation."""
        s = '\t'.join( [self.chrom, str(self.initial), str(self.final)] )
        # Name
        if self.name is not None: s += '\t' + self.name
        else: s += '\t' + self.toString()
        # Score
        if not self.data: score = "."
        else:
            data = self.data.split("\t")
            try: 
                score = data[0]
            except: score = "."
        s += '\t' + score
        # Orientation
        if self.orientation: s += '\t' + self.orientation
        else: s += '\t' + "."
        # Else
        if self.data:
            s += '\t' + "\t".join(data[1:])
        return s

    def __hash__(self):
        return hash(tuple([self.chrom, self.initial, self.final, self.orientation]))

    def __eq__(self, other):
        return (self.chrom, self.initial, self.final, self.orientation) == (other.chrom, other.initial, other.final, other.orientation)

    def toString(self, space=False):
        """Return a string of GenomicRegion by its position.

        *Keyword arguments:*

            - space -- insert spaces between the values.
        """
        if space:
            return "chr "+self.chrom[3:]+": "+str(self.initial)+"-"+str(self.final)
        else:
            return self.chrom+":"+str(self.initial)+"-"+str(self.final)

    def extend(self, left, right, w_return=False):
        """Extend GenomicRegion on both sides.

        *Keyword arguments:*
        
            - left -- Define the length to extend on left.
            - right -- Define the length to extend on right.
        """

        initial = self.initial - left
        final = self.final + right

        #if left, right are negative, switching the border may be necessary
        if initial > final:
            initial, final = final, initial

        initial = max(initial, 0)

        if w_return:
            return GenomicRegion(chrom=self.chrom,initial=initial, final=final,
                                 name=self.name, orientation=self.orientation, data=self.data)
        else:
            self.initial = initial
            self.final = final

    def overlap(self, region):
        """Return True, if GenomicRegion overlaps with region, else False.

        *Keyword arguments:*

            - region -- Given GenomicRegion to compare.
        """
        if region.chrom == self.chrom:
            if self.initial <= region.initial:
                if self.final > region.initial:
                    return True
                #elif self.initial == self.final:
                #    raise Exception(self.chrom+","+str(self.initial)+","+str(self.final)+"\tThe region length shouldn't be zero. Please extend the region.")
            else:
                if self.initial < region.final:
                    return True
        return False
                    
    def __repr__(self):
        """Return official representation of GenomicRegion."""
        return ','.join( [self.chrom, str(self.initial), str(self.final)] )

    def __cmp__(self, region):
        """Return negative value if x < y, zero if x == y and strictly positive if x > y.

        *Keyword arguments:*

            - region -- Given GenomicRegion to compare.
        """
        if self.chrom < region.chrom:
            return -1
        elif self.chrom > region.chrom:
            return 1
        else:
            if self.initial < region.initial:
                return -1
            elif self.initial > region.initial:
                return 1
            else:
                if self.final < region.final:
                    return -1
                elif self.final > region.final:
                    return 1
                else:
                    return 0

    def extract_blocks(self, keep_name=False):
        """Extract the block information in self.data into a GenomicRegionSet."""
        z = []
        data = self.data.split("\t")
        nexon = int(data[4])
        width = data[5].split(",")
        posit = data[6].split(",")
        reverse = False
        for i in range(nexon):
            if self.orientation == "-":
                n = "_exon_"+str(nexon - i)
                reverse = True
            else:
                n = "_exon_"+str(i + 1)
            if keep_name: name = self.name
            else: name = self.name+n
            z.append(GenomicRegion(chrom=self.chrom, 
                                   initial=self.initial+int(posit[i]), 
                                   final=self.initial+int(posit[i])+int(width[i]), 
                                   name=name, orientation=self.orientation, data=data[0] ))
        if reverse: return z[::-1]
        else: return z

    def distance(self, y):
        """Return the distance between two GenomicRegions. If overlapping, return 0; if on different chromosomes, return None.

        *Keyword arguments:*

            - y -- Given GenomicRegion to compare.
        """
        
        if self.chrom == y.chrom:
            if self.overlap(y):
                return 0
            elif self < y:
                return y.initial - self.final
            else:
                return self.initial - y.final
        else:
            return None




"""
GenomicVariant
===================
GenomicVariant is a specialized GenomicRegion class and describes a SNP or InDel.

"""


from rgt.GenomicRegion import GenomicRegion

class GenomicVariant(GenomicRegion):
    """*Keyword arguments:*

        - chrom -- chromosome
        - pos -- position
        - ref -- reference nucleotide
        - alt -- alternative nucleotide
        - qual -- quality
        - filter -- filter
        - id -- id
        - info -- further informaton
        - format -- SNP format
        - genotype -- genotype
        - samples -- sample
        
    .. note:: all necessary information are contained in a VCF file.
    """
    
    def __init__(self, chrom, pos, ref, alt, qual, filter = None, id = None, info = None, format = None, genotype = None, samples = None):
        GenomicRegion.__init__(self, chrom, pos, pos + 1)
        
        self.chrom = str(chrom)
        self.pos = int(pos)
        self.id = id
        self.ref = ref
        self.alt = alt
        self.qual = qual
        self.filter = filter 
        self.info = info
        self.format = format
        self.genotype = genotype
        self.samples = samples
        
        self.name = self.__str__
        self.data = "_$_".join(map(lambda x: str(x), [self.id, self.ref, self.alt, self.qual, self.filter, self.info, self.format, self.genotype, self.samples]))
    
    def __str__(self):
        """Give informal string representation"""
        s = '-'.join( [self.chrom, str(self.pos)] )
        if self.id is not None:
            s += '-' + self.id
        
        return s

    def __repr__(self):
        """Return official representation of GenomicRegion"""
        return ','.join( [self.chrom, str(self.pos)] )

if __name__ == '__main__':
    a = GenomicVariant('chr1', 1, "T", "A", 20)
    print(a)